@@ -48,71 +48,129 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
+#include "dkg-io.hh"
 #include "dkg-common.hh"
 
-int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-pid_t 					pid[DKGPG_MAX_N];
-std::vector<std::string>		peers;
-bool					instance_forked = false;
+int							pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+pid_t						pid[DKGPG_MAX_N];
+std::vector<std::string>	peers;
+bool						instance_forked = false;
 
-std::string				passphrase, userid, passwords, hostname, port;
-tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
-std::map<size_t, size_t>		idx2dkg, dkg2idx;
-mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
-size_t					dss_n, dss_t, dss_i;
-std::vector<size_t>			dss_qual, dss_x_rvss_qual;
-std::vector< std::vector<mpz_ptr> >	dss_c_ik;
-mpz_t					dkg_p, dkg_q, dkg_g, dkg_h, dkg_x_i, dkg_xprime_i, dkg_y;
-size_t					dkg_n, dkg_t, dkg_i;
-std::vector<size_t>			dkg_qual;
-std::vector<mpz_ptr>			dkg_v_i;
-std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
-gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
-gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
-gcry_mpi_t 				gk, myk, sig_r, sig_s;
-gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
+std::string					passphrase, kfilename;
+std::string					passwords, hostname, port;
 
-int 					opt_verbose = 0;
-bool					libgcrypt_secmem = false;
-char					*opt_passwords = NULL;
-char					*opt_hostname = NULL;
+int 						opt_verbose = 0;
+char						*opt_passwords = NULL;
+char						*opt_hostname = NULL;
+char						*opt_k = NULL;
 unsigned long int			opt_p = 55000, opt_W = 5;
 
-mpz_t					cache[TMCG_MAX_SSRANDOMM_CACHE], cache_mod;
-size_t					cache_avail = 0;
+mpz_t						cache[TMCG_MAX_SSRANDOMM_CACHE], cache_mod;
+size_t						cache_avail = 0;
 
 void run_instance
 	(size_t whoami, const size_t num_xtests)
 {
-	// read and parse the private key
+	// read the key file
 	std::string armored_seckey, thispeer = peers[whoami];
+	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
+	{
+		std::cerr << "WARNING: weak permissions of private key file" <<
+			" detected" << std::endl;
+		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
+			exit(-1);
+	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
-	init_mpis();
-	std::vector<std::string> CAPL;
-	time_t ckeytime = 0, ekeytime = 0;
-	if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
+
+	// read the keyring
+	std::string armored_pubring;
+	if (opt_k)
+	{
+		if (!read_key_file(kfilename, armored_pubring))
+			exit(-1);
+	}
+
+	// parse the keyring, the private key and corresponding signatures
+	TMCG_OpenPGP_Prvkey *prv = NULL;
+	TMCG_OpenPGP_Keyring *ring = NULL;
+	bool parse_ok;
+	if (opt_k)
+	{
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+		if (!parse_ok)
+		{
+			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
+			ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
+		}
+	}
+	else
+		ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
+	parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+		PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
+	if (!parse_ok)
 	{
-		release_mpis();
-		keyid.clear(), subkeyid.clear(), pub.clear(), sub.clear(), uidsig.clear(), subsig.clear();
-		dss_qual.clear(), dss_x_rvss_qual.clear(), dss_c_ik.clear(), dkg_qual.clear(), dkg_v_i.clear(), dkg_c_ik.clear();
-		init_mpis();
-		// protected with password
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		if (!get_passphrase("Enter passphrase to unlock private key",
+			passphrase))
 		{
-			release_mpis();
+			std::cerr << "ERROR: cannot read passphrase" << std::endl;
+			delete ring;
 			exit(-1);
 		}
 #endif
-		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
-		{
-			std::cerr << "R_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
-			release_mpis();
-			exit(-1);
-		}
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
+	}
+	if (parse_ok)
+	{
+		prv->RelinkPublicSubkeys(); // relink the contained subkeys
+		prv->pub->CheckSelfSignatures(ring, opt_verbose);
+		prv->pub->CheckSubkeys(ring, opt_verbose);
+		prv->RelinkPrivateSubkeys(); // undo the relinking
+	}
+	else
+	{
+		std::cerr << "ERROR: cannot use the provided private key" << std::endl;
+		delete ring;
+		exit(-1);
+	}
+	if (!prv->pub->valid || prv->weak(opt_verbose))
+	{
+		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	if (prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7)
+	{
+		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+
+	// create an instance of tDSS by stored parameters from private key
+	CanettiGennaroJareckiKrawczykRabinDSS *dss = NULL;
+	if (!init_tDSS(prv, opt_verbose, dss))
+	{
+		delete dss;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	// create one-to-one mapping based on the stored canonicalized peer list
+	if (!prv->tDSS_CreateMapping(peers, opt_verbose))
+	{
+		std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" << std::endl;
+		delete dss;
+		delete ring;
+		delete prv;
+		exit(-1);
 	}
 
 	// create communication handles between all players
@@ -126,20 +184,31 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "R_" << whoami << ": " << "cannot read password for protecting channel to R_" << i << std::endl;
-				release_mpis();
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read" <<
+					" password for protecting channel to R_" << i << std::endl;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1);
 			}
 			key << pwd;
-			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
+			if (((i + 1) < peers.size()) &&
+				!TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "R_" << whoami << ": " << "cannot skip to next password for protecting channel to R_" << (i + 1) << std::endl;
-				release_mpis();
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip" <<
+					" to next password for protecting channel to R_" <<
+					(i + 1) << std::endl;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1);
 			}
 		}
 		else
-			key << "dkg-refresh::R_" << (i + whoami); // use simple key -- we assume that GNUnet will provide secure channels
+		{
+			// use simple key -- we assume that GNUnet provides secure channels
+			key << "dkg-refresh::R_" << (i + whoami);
+		}
 		uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
 		uP_key.push_back(key.str());
@@ -149,17 +218,24 @@ void run_instance
 	}
 
 	// create asynchronous authenticated unicast channels
-	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami,
+		uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin,
+		(opt_W * 60));
 
 	// create asynchronous authenticated unicast channels for broadcast protocol
-	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami,
+		bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin,
+		(opt_W * 60));
 			
 	// create an instance of a reliable broadcast protocol (RBC)
 	std::string myID = "dkg-refresh|";
 	for (size_t i = 0; i < peers.size(); i++)
 		myID += peers[i] + "|";
-	size_t T_RBC = (peers.size() - 1) / 3; // assume maximum asynchronous t-resilience for RBC
-	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	// assume maximum asynchronous t-resilience for RBC
+	size_t T_RBC = (peers.size() - 1) / 3;
+	CachinKursawePetzoldShoupRBC *rbc =
+		new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2,
+			aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
@@ -167,181 +243,73 @@ void run_instance
 	{
 		mpz_t xtest;
 		mpz_init_set_ui(xtest, i);
-		std::cout << "R_" << whoami << ": xtest = " << xtest << " <-> ";
+		std::cerr << "INFO: R_" << whoami << ": xtest = " << xtest << " <-> ";
 		rbc->Broadcast(xtest);
 		for (size_t ii = 0; ii < peers.size(); ii++)
 		{
 			if (!rbc->DeliverFrom(xtest, ii))
-				std::cout << "<X> ";
+				std::cerr << "<X> ";
 			else
-				std::cout << xtest << " ";
+				std::cerr << xtest << " ";
 		}
-		std::cout << std::endl;
+		std::cerr << std::endl;
 		mpz_clear(xtest);
 	}
 
-	// create an instance of tDSS by stored parameters from private key
-	std::stringstream dss_in;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl << dss_y << std::endl;
-	dss_in << dss_qual.size() << std::endl;
-	for (size_t i = 0; i < dss_qual.size(); i++)
-		dss_in << dss_qual[i] << std::endl;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl << dss_y << std::endl;
-	dss_in << dss_qual.size() << std::endl;
-	for (size_t i = 0; i < dss_qual.size(); i++)
-		dss_in << dss_qual[i] << std::endl;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl << dss_t << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl;
-	dss_in << "0" << std::endl << "0" << std::endl;
-	if (dss_x_rvss_qual.size())
-	{
-		// new private key format 107
-		dss_in << dss_x_rvss_qual.size() << std::endl;
-		for (size_t i = 0; i < dss_x_rvss_qual.size(); i++)
-			dss_in << dss_x_rvss_qual[i] << std::endl;
-	}
-	else
-	{
-		// old private key format 108
-		dss_in << dss_qual.size() << std::endl;
-		for (size_t i = 0; i < dss_qual.size(); i++)
-			dss_in << dss_qual[i] << std::endl;
-	}
-	assert((dss_c_ik.size() == dss_n));
-	for (size_t i = 0; i < dss_c_ik.size(); i++)
-	{
-		for (size_t j = 0; j < dss_c_ik.size(); j++)
-			dss_in << "0" << std::endl << "0" << std::endl;
-		assert((dss_c_ik[i].size() == (dss_t + 1)));
-		for (size_t k = 0; k < dss_c_ik[i].size(); k++)
-			dss_in << dss_c_ik[i][k] << std::endl;
-	}
-	if (opt_verbose)
-		std::cout << "CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
-	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
-	if (!dss->CheckGroup())
-	{
-		std::cerr << "R_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	// update the tDSS keys (proactive security against mobile adversary)
+	// update the tDSS key (proactive security against mobile adversary)
 	std::stringstream err_log;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": dss.Refresh()" << std::endl;
-	if (!dss->Refresh(peers.size(), whoami, idx2dkg, dkg2idx, aiou, rbc, err_log, false, cache, cache_mod, &cache_avail))
+		std::cerr << "INFO: R_" << whoami << ": dss.Refresh()" << std::endl;
+	if (!dss->Refresh(peers.size(), whoami, prv->tdss_idx2dkg,
+		prv->tdss_dkg2idx, aiou, rbc, err_log, false, cache, cache_mod,
+		&cache_avail))
 	{
-		std::cerr << "R_" << whoami << ": " << "tDSS Refresh() failed" << std::endl;
-		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log.str();
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Refresh() failed" <<
+			std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl <<
+			err_log.str();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cout << "R_" << whoami << ": log follows " << std::endl << err_log.str();
+		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log.str();
 	// at the end: deliver some more rounds for still waiting parties
-	time_t synctime = aiounicast::aio_timeout_very_long;
+	time_t synctime = aiounicast::aio_timeout_long;
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": waiting approximately " <<
+			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
+			std::endl;
 	rbc->Sync(synctime);
 	// create an OpenPGP DSA-based primary key using refreshed values from tDSS
-	gcry_mpi_t p, q, g, h, y, n, t, i, qualsize, x_rvss_qualsize, x_i, xprime_i;
+	gcry_mpi_t n, t, i, qualsize, x_rvss_qualsize, x_i, xprime_i;
 	std::vector<gcry_mpi_t> qual, x_rvss_qual;
 	std::vector< std::vector<gcry_mpi_t> > c_ik;
-	p = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(p, dss->p))
+	x_i = gcry_mpi_snew(2048);
+	if (!tmcg_mpz_get_gcry_mpi(x_i, dss->x_i))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for p" << std::endl;
-		gcry_mpi_release(p);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	q = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(q, dss->q))
-	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for q" << std::endl;
-		gcry_mpi_release(p);
-		gcry_mpi_release(q);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	g = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(g, dss->g))
-	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for g" << std::endl;
-		gcry_mpi_release(p);
-		gcry_mpi_release(q);
-		gcry_mpi_release(g);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	h = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(h, dss->h))
-	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for h" << std::endl;
-		gcry_mpi_release(p);
-		gcry_mpi_release(q);
-		gcry_mpi_release(g);
-		gcry_mpi_release(h);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	y = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(y, dss->y))
-	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for y" << std::endl;
-		gcry_mpi_release(p);
-		gcry_mpi_release(q);
-		gcry_mpi_release(g);
-		gcry_mpi_release(h);
-		gcry_mpi_release(y);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	if (libgcrypt_secmem)
-		x_i = gcry_mpi_snew(2048);
-	else
-		x_i = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(x_i, dss->x_i))
-	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for x_i" << std::endl;
-		gcry_mpi_release(p);
-		gcry_mpi_release(q);
-		gcry_mpi_release(g);
-		gcry_mpi_release(h);
-		gcry_mpi_release(y);
+		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+			" failed for x_i" << std::endl;
 		gcry_mpi_release(x_i);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	if (libgcrypt_secmem)
-		xprime_i = gcry_mpi_snew(2048);
-	else
-		xprime_i = gcry_mpi_new(2048);
-	if (!mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
+	xprime_i = gcry_mpi_snew(2048);
+	if (!tmcg_mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
 	{
-		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for xprime_i" << std::endl;
-		gcry_mpi_release(p);
-		gcry_mpi_release(q);
-		gcry_mpi_release(g);
-		gcry_mpi_release(h);
-		gcry_mpi_release(y);
+		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+			" failed for xprime_i" << std::endl;
 		gcry_mpi_release(x_i);
 		gcry_mpi_release(xprime_i);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	n = gcry_mpi_set_ui(NULL, dss->n);
@@ -366,14 +334,11 @@ void run_instance
 		{
 			gcry_mpi_t tmp;
 			tmp = gcry_mpi_new(2048);
-			if (!mpz_get_gcry_mpi(tmp, dss->dkg->x_rvss->C_ik[j][k]))
+			if (!tmcg_mpz_get_gcry_mpi(tmp, dss->dkg->x_rvss->C_ik[j][k]))
 			{
-				std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dss->dkg->x_rvss->C_ik[j][k]" << std::endl;
-				gcry_mpi_release(p);
-				gcry_mpi_release(q);
-				gcry_mpi_release(g);
-				gcry_mpi_release(h);
-				gcry_mpi_release(y);
+				std::cerr << "ERROR: R_" << whoami <<
+					": tmcg_mpz_get_gcry_mpi() failed for" <<
+					" dss->dkg->x_rvss->C_ik[j][k]" << std::endl;
 				gcry_mpi_release(x_i);
 				gcry_mpi_release(xprime_i);
 				gcry_mpi_release(n);
@@ -389,21 +354,21 @@ void run_instance
 					for (size_t kk = 0; kk < c_ik[jj].size(); kk++)
 						gcry_mpi_release(c_ik[jj][kk]);
 				gcry_mpi_release(tmp);
-				delete dss, delete rbc, delete aiou, delete aiou2;
-				release_mpis();
+				delete rbc, delete aiou, delete aiou2;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1); 
 			}
 			c_ik[j].push_back(tmp);
 		}
 	}
-	sec.clear();
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSecEncodeExperimental107(ckeytime, p, q, g, h, y, 
-		n, t, i, qualsize, qual, x_rvss_qualsize, x_rvss_qual, CAPL, c_ik, x_i, xprime_i, passphrase, sec);
-	gcry_mpi_release(p);
-	gcry_mpi_release(q);
-	gcry_mpi_release(g);
-	gcry_mpi_release(h);
-	gcry_mpi_release(y);
+	tmcg_openpgp_octets_t sec;
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		PacketSecEncodeExperimental107(prv->pub->creationtime, prv->pub->dsa_p,
+			prv->pub->dsa_q, prv->pub->dsa_g, prv->tdss_h, prv->pub->dsa_y, n,
+			t, i, qualsize, qual, x_rvss_qualsize, x_rvss_qual, prv->tdss_capl,
+			c_ik, x_i, xprime_i, passphrase, sec);
 	gcry_mpi_release(x_i);
 	gcry_mpi_release(xprime_i);
 	gcry_mpi_release(n);
@@ -419,66 +384,138 @@ void run_instance
 		for (size_t k = 0; k < c_ik[j].size(); k++)
 			gcry_mpi_release(c_ik[j][k]);
 	
-	// export updated private keys in OpenPGP armor format
-	tmcg_octets_t all;
+	// export updated private key in OpenPGP armor format
+	tmcg_openpgp_octets_t all;
 	std::string armor;
 	std::stringstream secfilename;
 	secfilename << peers[whoami] << "_dkg-sec.asc";
 	all.insert(all.end(), sec.begin(), sec.end());
-	all.insert(all.end(), uid.begin(), uid.end());
-	all.insert(all.end(), uidsig.begin(), uidsig.end());
-	all.insert(all.end(), ssb.begin(), ssb.end());
-	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
+	for (size_t k = 0; k < prv->pub->selfsigs.size(); k++)
+		all.insert(all.end(),
+			(prv->pub->selfsigs[k]->packet).begin(),
+			(prv->pub->selfsigs[k]->packet).end());
+	for (size_t k = 0; k < prv->pub->keyrevsigs.size(); k++)
+		all.insert(all.end(),
+			(prv->pub->keyrevsigs[k]->packet).begin(),
+			(prv->pub->keyrevsigs[k]->packet).end());
+	for (size_t j = 0; j < prv->pub->userids.size(); j++)
+	{
+		TMCG_OpenPGP_UserID *uid = prv->pub->userids[j];
+		if (uid->valid)
+		{
+			all.insert(all.end(),
+				(uid->packet).begin(), (uid->packet).end());
+			for (size_t k = 0; k < uid->selfsigs.size(); k++)
+				all.insert(all.end(),
+					(uid->selfsigs[k]->packet).begin(),
+					(uid->selfsigs[k]->packet).end());
+			for (size_t k = 0; k < uid->revsigs.size(); k++)
+				all.insert(all.end(),
+					(uid->selfsigs[k]->packet).begin(),
+					(uid->selfsigs[k]->packet).end());
+		}
+	}
+	for (size_t j = 0; j < prv->pub->userattributes.size(); j++)
+	{
+		TMCG_OpenPGP_UserAttribute *uat = prv->pub->userattributes[j];
+		if (uat->valid)
+		{
+			all.insert(all.end(),
+				(uat->packet).begin(), (uat->packet).end());
+			for (size_t k = 0; k < uat->selfsigs.size(); k++)
+				all.insert(all.end(),
+					(uat->selfsigs[k]->packet).begin(),
+					(uat->selfsigs[k]->packet).end());
+			for (size_t k = 0; k < uat->revsigs.size(); k++)
+				all.insert(all.end(),
+					(uat->selfsigs[k]->packet).begin(),
+					(uat->selfsigs[k]->packet).end());
+		}
+	}
+	for (size_t j = 0; j < prv->private_subkeys.size(); j++)
+	{
+		TMCG_OpenPGP_PrivateSubkey *ssb = prv->private_subkeys[j];
+		all.insert(all.end(), (ssb->packet).begin(), (ssb->packet).end());
+		TMCG_OpenPGP_Subkey *sub = prv->private_subkeys[j]->pub;
+		for (size_t k = 0; k < sub->selfsigs.size(); k++)
+			all.insert(all.end(),
+				(sub->selfsigs[k]->packet).begin(),
+				(sub->selfsigs[k]->packet).end());
+		for (size_t k = 0; k < sub->bindsigs.size(); k++)
+			all.insert(all.end(),
+				(sub->bindsigs[k]->packet).begin(),
+				(sub->bindsigs[k]->packet).end());
+		for (size_t k = 0; k < sub->pbindsigs.size(); k++)
+			all.insert(all.end(),
+				(sub->pbindsigs[k]->packet).begin(),
+				(sub->pbindsigs[k]->packet).end());
+		for (size_t k = 0; k < sub->keyrevsigs.size(); k++)
+			all.insert(all.end(),
+				(sub->keyrevsigs[k]->packet).begin(),
+				(sub->keyrevsigs[k]->packet).end());
+	}
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		ArmorEncode(TMCG_OPENPGP_ARMOR_PRIVATE_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
-	std::ofstream secofs((secfilename.str()).c_str(), std::ofstream::out | std::ofstream::trunc);
+	std::ofstream secofs((secfilename.str()).c_str(),
+		std::ofstream::out | std::ofstream::trunc);
 	if (!secofs.good())
 	{
-		std::cerr << "R_" << whoami << ": opening private key file failed" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		std::cerr << "ERROR: R_" << whoami << ": opening private key file failed" << std::endl;
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	secofs << armor;
 	if (!secofs.good())
 	{
-		std::cerr << "R_" << whoami << ": writing private key file failed" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		std::cerr << "ERROR: R_" << whoami << ": writing private key file failed" << std::endl;
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	secofs.close();
 
-	// release tDSS
-	delete dss;
-
 	// release RBC
 	delete rbc;
 	
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": aiou.numRead = " << aiou->numRead <<
-			" aiou.numWrite = " << aiou->numWrite << std::endl;
+	{
+		std::cerr << "INFO: R_" << whoami << ": unicast channels";
+		aiou->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cout << "R_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
-			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
+	{
+		std::cerr << "INFO: R_" << whoami << ": broadcast channel";
+		aiou2->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
 
 	// release asynchronous unicast and broadcast
 	delete aiou, delete aiou2;
 
 	// release
-	release_mpis();
+	delete dss;
+	delete ring;
+	delete prv;
 }
 
 #ifdef GNUNET
 char *gnunet_opt_hostname = NULL;
 char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
+char *gnunet_opt_k = NULL;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
 unsigned int gnunet_opt_W = opt_W;
@@ -489,7 +526,7 @@ void fork_instance
 	(const size_t whoami)
 {
 	if ((pid[whoami] = fork()) < 0)
-		perror("dkg-refresh (fork)");
+		perror("ERROR: dkg-refresh (fork)");
 	else
 	{
 		if (pid[whoami] == 0)
@@ -501,14 +538,14 @@ void fork_instance
 			run_instance(whoami, 0);
 #endif
 			if (opt_verbose)
-				std::cout << "R_" << whoami << ": exit(0)" << std::endl;
+				std::cerr << "INFO: R_" << whoami << ": exit(0)" << std::endl;
 			exit(0);
 			/* END child code: participant R_i */
 		}
 		else
 		{
 			if (opt_verbose)
-				std::cout << "fork() = " << pid[whoami] << std::endl;
+				std::cerr << "INFO: fork() = " << pid[whoami] << std::endl;
 			instance_forked = true;
 		}
 	}
@@ -531,6 +568,12 @@ int main
 			"hostname (e.g. onion address) of this peer within PEERS",
 			&gnunet_opt_hostname
 		),
+		GNUNET_GETOPT_option_string('k',
+			"keyring",
+			"FILENAME",
+			"use keyring FILENAME containing external revocation keys",
+			&gnunet_opt_k
+		),
 		GNUNET_GETOPT_option_logfile(&logfile),
 		GNUNET_GETOPT_option_loglevel(&loglev),
 		GNUNET_GETOPT_option_string('p',
@@ -589,6 +632,8 @@ int main
 		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
+	if (gnunet_opt_k != NULL)
+		opt_k = gnunet_opt_k;
 	if (gnunet_opt_W != opt_W)
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 #endif
@@ -598,8 +643,12 @@ int main
 	{
 		std::string arg = argv[i+1];
 		// ignore options
-		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) || (arg.find("-r") == 0) || (arg.find("-w") == 0) || (arg.find("-L") == 0) || 
-			(arg.find("-l") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0) || (arg.find("-W") == 0))
+		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) ||
+			(arg.find("-r") == 0) || (arg.find("-w") == 0) ||
+			(arg.find("-L") == 0) || (arg.find("-l") == 0) ||
+			(arg.find("-x") == 0) || (arg.find("-P") == 0) ||
+			(arg.find("-H") == 0) || (arg.find("-W") == 0) ||
+			(arg.find("-k") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) && (opt_hostname == NULL))
@@ -607,6 +656,12 @@ int main
 				hostname = argv[i+1];
 				opt_hostname = (char*)hostname.c_str();
 			}
+			if ((arg.find("-k") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_k == NULL))
+			{
+				kfilename = argv[i+1];
+				opt_k = (char*)kfilename.c_str();
+			}
 			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) && (opt_passwords == NULL))
 			{
 				passwords = argv[i+1];
@@ -628,6 +683,8 @@ int main
 				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
 				std::cout << "  -H STRING      hostname (e.g. onion address) of this peer within PEERS" << std::endl;
+				std::cout << "  -k FILENAME    use keyring FILENAME" <<
+					" containing external revocation keys" << std::endl;
 				std::cout << "  -p INTEGER     start port for built-in TCP/IP message exchange service" << std::endl;
 				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
 				std::cout << "  -v, --version  print the version number" << std::endl;
@@ -667,7 +724,7 @@ int main
 	peers.push_back("Test1");
 	peers.push_back("Test4");
 	peers.push_back("Test3");
-	opt_verbose = 1;
+	opt_verbose = 2;
 #endif
 
 	// check command line arguments
@@ -693,9 +750,9 @@ int main
 	}
 	if (opt_verbose)
 	{
-		std::cout << "INFO: canonicalized peer list = " << std::endl;
+		std::cerr << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
-			std::cout << peers[i] << std::endl;
+			std::cerr << peers[i] << std::endl;
 	}
 
 	// lock memory
@@ -713,7 +770,6 @@ int main
 		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
 		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
 		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
-		libgcrypt_secmem = true;
 	}
 
 	// initialize LibTMCG
@@ -723,9 +779,9 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
 
-	// initialize cache
+	// read the public key
 	std::string armored_pubkey = "undefined";
 	if (opt_hostname != NULL)
 	{
@@ -745,34 +801,43 @@ int main
 		std::cerr << "ERROR: no corresponding public key file found" << std::endl;
 		return -1;
 	}
-	init_mpis();
-	time_t ckeytime = 0, ekeytime = 0, csubkeytime = 0, esubkeytime = 0;
-	tmcg_byte_t keyusage = 0, keystrength = 1;
-	if (!parse_public_key(armored_pubkey, ckeytime, ekeytime, csubkeytime, esubkeytime, keyusage, keystrength))
+
+	// parse the public key block and corresponding signatures
+	TMCG_OpenPGP_Pubkey *primary = NULL;
+	bool parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+		PublicKeyBlockParse(armored_pubkey, opt_verbose, primary);
+	if (parse_ok)
 	{
-		std::cerr << "ERROR: cannot parse the provided public key" << std::endl;
-		release_mpis();
+		if (primary->pkalgo != TMCG_OPENPGP_PKALGO_DSA)
+		{
+			std::cerr << "ERROR: primary key is not DSA" << std::endl;
+			delete primary;
+			return -1;
+		}
+	}
+	else
+	{
+		std::cerr << "ERROR: cannot use the provided public key" << std::endl;
 		return -1;
 	}
-	if (!keystrength)
+
+	// initialize cache
+	mpz_t dss_q;
+	mpz_init(dss_q);
+	if (!tmcg_mpz_set_gcry_mpi(primary->dsa_q, dss_q))
 	{
-		std::cerr << "ERROR: provided public key is too weak" << std::endl;
-		release_mpis();
+		std::cerr << "ERROR: tmcg_mpz_set_gcry_mpi() failed for dss_q" <<
+			std::endl;
+		mpz_clear(dss_q);
+		delete primary;
 		return -1;
 	}
-	if (!mpz_set_gcry_mpi(dsa_q, dss_q))
-	{
-		std::cerr << "ERROR: mpz_set_gcry_mpi() failed for dss_q" << std::endl;
-		release_mpis();
-		return -1;
-	}
-	std::cout << "We need some entropy to cache very strong randomness for share refresh." << std::endl;
-	std::cout << "Please use other programs, move the mouse, and type on your keyboard: " << std::endl; 
-	mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail, (2 * peers.size()), dss_q);
-	std::cout << "Thank you!" << std::endl;
-	keyid.clear(), subkeyid.clear(), pub.clear(), sub.clear(), uidsig.clear(), subsig.clear();
-	dss_qual.clear(), dss_x_rvss_qual.clear(), dss_c_ik.clear(), dkg_qual.clear(), dkg_v_i.clear(), dkg_c_ik.clear();
-	release_mpis();
+	delete primary;
+	std::cerr << "We need some entropy to cache very strong randomness for share refresh." << std::endl;
+	std::cerr << "Please use other programs, move the mouse, and type on your keyboard: " << std::endl; 
+	tmcg_mpz_ssrandomm_cache_init(cache, cache_mod, &cache_avail, (2 * peers.size()), dss_q);
+	std::cerr << "Thank you!" << std::endl;
+	mpz_clear(dss_q);
 
 	// initialize return code
 	int ret = 0;
@@ -781,6 +846,11 @@ int main
 	{
 		if (port.length())
 			opt_p = strtoul(port.c_str(), NULL, 10); // get start port from options
+		if ((opt_p < 1) || (opt_p > 65535))
+		{
+			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
+			return -1;
+		}
 		tcpip_init(hostname);
 		tcpip_bindports((uint16_t)opt_p, false);
 		tcpip_bindports((uint16_t)opt_p, true);
@@ -794,7 +864,7 @@ int main
 		tcpip_close();
 		tcpip_done();
 		// release cache
-		mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+		tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 		return ret;
 	}
 
@@ -807,6 +877,12 @@ int main
 			"hostname (e.g. onion address) of this peer within PEERS",
 			&gnunet_opt_hostname
 		),
+		GNUNET_GETOPT_option_string('k',
+			"keyring",
+			"FILENAME",
+			"use keyring FILENAME containing external revocation keys",
+			&gnunet_opt_k
+		),
 		GNUNET_GETOPT_option_string('p',
 			"port",
 			"STRING",
@@ -847,7 +923,7 @@ int main
 	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run, argv[0]);
 	GNUNET_free((void *) argv);
 	// release cache
-	mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 	if (ret == GNUNET_OK)
 		return 0;
 	else
@@ -856,7 +932,7 @@ int main
 	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
 #endif
 
-	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
 	std::cerr << "WARNING: due to cache issues the generated shares are identical, don't use them!" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
@@ -864,9 +940,9 @@ int main
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if (pipe(pipefd[i][j]) < 0)
-				perror("dkg-refresh (pipe)");
+				perror("ERROR: dkg-refresh (pipe)");
 			if (pipe(broadcast_pipefd[i][j]) < 0)
-				perror("dkg-refresh (pipe)");
+				perror("ERROR: dkg-refresh (pipe)");
 		}
 	}
 	
@@ -882,9 +958,9 @@ int main
 	{
 		int wstatus = 0;
 		if (opt_verbose)
-			std::cout << "waitpid(" << pid[i] << ")" << std::endl;
+			std::cerr << "INFO: waitpid(" << pid[i] << ")" << std::endl;
 		if (waitpid(pid[i], &wstatus, 0) != pid[i])
-			perror("dkg-refresh (waitpid)");
+			perror("ERROR: dkg-refresh (waitpid)");
 		if (!WIFEXITED(wstatus))
 		{
 			std::cerr << "ERROR: protocol instance ";
@@ -897,21 +973,21 @@ int main
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cout << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("dkg-refresh (close)");
+				perror("ERROR: dkg-refresh (close)");
 			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("dkg-refresh (close)");
+				perror("ERROR: dkg-refresh (close)");
 		}
 	}
 	
 	// release cache
-	mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
+	tmcg_mpz_ssrandomm_cache_done(cache, cache_mod, &cache_avail);
 	// finish	
 	return ret;
 }
