@@ -48,75 +48,130 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
-#include "dkg-common.hh"
 #include "dkg-io.hh"
+#include "dkg-common.hh"
 
-int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-pid_t 					pid[DKGPG_MAX_N];
-std::vector<std::string>		peers;
-bool					instance_forked = false;
+int 						pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+pid_t 						pid[DKGPG_MAX_N];
+std::vector<std::string>	peers;
+bool						instance_forked = false;
 
-std::string				passphrase, userid, passwords, hostname, port;
-tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
-std::map<size_t, size_t>		idx2dkg, dkg2idx;
-mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
-size_t					dss_n, dss_t, dss_i;
-std::vector<size_t>			dss_qual, dss_x_rvss_qual;
-std::vector< std::vector<mpz_ptr> >	dss_c_ik;
-mpz_t					dkg_p, dkg_q, dkg_g, dkg_h, dkg_x_i, dkg_xprime_i, dkg_y;
-size_t					dkg_n, dkg_t, dkg_i;
-std::vector<size_t>			dkg_qual;
-std::vector<mpz_ptr>			dkg_v_i;
-std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
-gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
-gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
-gcry_mpi_t 				gk, myk, sig_r, sig_s;
-gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
+std::string					passphrase, kfilename;
+std::string					passwords, hostname, port;
 
-int 					opt_verbose = 0;
-bool					libgcrypt_secmem = false;
-char					*opt_passwords = NULL;
-char					*opt_hostname = NULL;
+int 						opt_verbose = 0;
+char						*opt_passwords = NULL;
+char						*opt_hostname = NULL;
+char						*opt_k = NULL;
 unsigned long int			opt_r = 0, opt_p = 55000, opt_W = 5;
 
 void run_instance
-	(size_t whoami, const time_t sigtime, const size_t reason, const size_t num_xtests)
+	(size_t whoami, const time_t sigtime, const size_t reason,
+	 const size_t num_xtests)
 {
-	// read and parse the private key
+	// read the key file
 	std::string armored_seckey, thispeer = peers[whoami];
 	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
 	{
-		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		std::cerr << "WARNING: weak permissions of private key file" <<
+			" detected" << std::endl;
 		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
 			exit(-1);
 	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
-	init_mpis();
-	std::vector<std::string> CAPL;
-	time_t ckeytime = 0, ekeytime = 0;
-	if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
+
+	// read the keyring
+	std::string armored_pubring;
+	if (opt_k)
+	{
+		if (!read_key_file(kfilename, armored_pubring))
+			exit(-1);
+	}
+
+	// parse the keyring, the private key and corresponding signatures
+	TMCG_OpenPGP_Prvkey *prv = NULL;
+	TMCG_OpenPGP_Keyring *ring = NULL;
+	bool parse_ok;
+	if (opt_k)
+	{
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+		if (!parse_ok)
+		{
+			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
+			ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
+		}
+	}
+	else
+		ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
+	parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+		PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
+	if (!parse_ok)
 	{
-		release_mpis();
-		keyid.clear(), subkeyid.clear(), pub.clear(), sub.clear(), uidsig.clear(), subsig.clear();
-		dss_qual.clear(), dss_x_rvss_qual.clear(), dss_c_ik.clear(), dkg_qual.clear(), dkg_v_i.clear(), dkg_c_ik.clear();
-		init_mpis();
-		// protected with password
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		if (!get_passphrase("Enter passphrase to unlock private key",
+			passphrase))
 		{
-			release_mpis();
+			std::cerr << "ERROR: cannot read passphrase" << std::endl;
+			delete ring;
 			exit(-1);
 		}
 #endif
-		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
-		{
-			std::cerr << "ERROR: R_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
-			release_mpis();
-			exit(-1);
-		}
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
+	}
+	if (parse_ok)
+	{
+		prv->RelinkPublicSubkeys(); // relink the contained subkeys
+		prv->pub->CheckSelfSignatures(ring, opt_verbose);
+		prv->pub->CheckSubkeys(ring, opt_verbose);
+		prv->RelinkPrivateSubkeys(); // undo the relinking
+	}
+	else
+	{
+		std::cerr << "ERROR: cannot use the provided private key" << std::endl;
+		delete ring;
+		exit(-1);
+	}
+	if (!prv->pub->valid || prv->weak(opt_verbose))
+	{
+		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	if (prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7)
+	{
+		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	TMCG_OpenPGP_PrivateSubkey *sub = NULL;
+	if (prv->private_subkeys.size())
+		sub = prv->private_subkeys[0];
+
+	// create an instance of tDSS by stored parameters from private key
+	CanettiGennaroJareckiKrawczykRabinDSS *dss = NULL;
+	if (!init_tDSS(prv, opt_verbose, dss))
+	{
+		delete dss;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	// create one-to-one mapping based on the stored canonicalized peer list
+	if (!prv->tDSS_CreateMapping(peers, opt_verbose))
+	{
+		std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" << std::endl;
+		delete dss;
+		delete ring;
+		delete prv;
+		exit(-1);
 	}
 
 	// create communication handles between all players
@@ -130,20 +185,31 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read password for protecting channel to R_" << i << std::endl;
-				release_mpis();
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot read" <<
+					" password for protecting channel to R_" << i << std::endl;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1);
 			}
 			key << pwd;
-			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
+			if (((i + 1) < peers.size()) &&
+				!TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip to next password for protecting channel to R_" << (i + 1) << std::endl;
-				release_mpis();
+				std::cerr << "ERROR: R_" << whoami << ": " << "cannot skip" <<
+					" to next password for protecting channel to R_" <<
+					(i + 1) << std::endl;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1);
 			}
 		}
 		else
-			key << "dkg-revoke::R_" << (i + whoami); // use simple key -- we assume that GNUnet will provide secure channels
+		{
+			// use simple key -- we assume that GNUnet provides secure channels
+			key << "dkg-revoke::R_" << (i + whoami);
+		}
 		uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
 		uP_key.push_back(key.str());
@@ -153,17 +219,24 @@ void run_instance
 	}
 
 	// create asynchronous authenticated unicast channels
-	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami,
+		uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin,
+		(opt_W * 60));
 
 	// create asynchronous authenticated unicast channels for broadcast protocol
-	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami,
+		bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin,
+		(opt_W * 60));
 			
 	// create an instance of a reliable broadcast protocol (RBC)
 	std::string myID = "dkg-revoke|";
 	for (size_t i = 0; i < peers.size(); i++)
 		myID += peers[i] + "|";
-	size_t T_RBC = (peers.size() - 1) / 3; // assume maximum asynchronous t-resilience for RBC
-	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	// assume maximum asynchronous t-resilience for RBC
+	size_t T_RBC = (peers.size() - 1) / 3;
+	CachinKursawePetzoldShoupRBC *rbc =
+		new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2,
+			aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
@@ -186,7 +259,8 @@ void run_instance
 
 	// participants must agree on a common signature creation time (OpenPGP)
 	if (opt_verbose)
-		std::cerr << "INFO: agree on a signature creation time for OpenPGP" << std::endl;
+		std::cerr << "INFO: agree on a signature creation time for OpenPGP" <<
+			std::endl;
 	time_t csigtime = 0;
 	std::vector<time_t> tvs;
 	mpz_t mtv;
@@ -205,7 +279,8 @@ void run_instance
 			}
 			else
 			{
-				std::cerr << "WARNING: R_" << whoami << ": no signature creation time stamp received from " << i << std::endl;
+				std::cerr << "WARNING: R_" << whoami << ": no signature" <<
+					" creation time stamp received from " << i << std::endl;
 			}
 		}
 	}
@@ -213,102 +288,69 @@ void run_instance
 	std::sort(tvs.begin(), tvs.end());
 	if (tvs.size() < (peers.size() - T_RBC))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": not enough timestamps received" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": not enough timestamps" <<
+			" received" << std::endl;
 		delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	csigtime = tvs[tvs.size()/2]; // use a median value as some kind of gentle agreement
+	// use a median value as some kind of gentle agreement
+	csigtime = tvs[tvs.size()/2];
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": canonicalized signature creation time = " << csigtime << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": canonicalized signature" <<
+			" creation time = " << csigtime << std::endl;
 
 	// select hash algorithm for OpenPGP based on |q| (size in bit)
 	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
-	if (mpz_sizeinbase(dss_q, 2L) == 256)
+	if (mpz_sizeinbase(dss->q, 2L) == 256)
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-	else if (mpz_sizeinbase(dss_q, 2L) == 384)
+	else if (mpz_sizeinbase(dss->q, 2L) == 384)
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-	else if (mpz_sizeinbase(dss_q, 2L) == 512)
+	else if (mpz_sizeinbase(dss->q, 2L) == 512)
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 	else
 	{
-		std::cerr << "ERROR: R_" << whoami << ": selecting hash algorithm failed for |q| = " << mpz_sizeinbase(dss_q, 2L) << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": selecting hash algorithm" <<
+			" failed for |q| = " << mpz_sizeinbase(dss->q, 2L) << std::endl;
 		delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 
-	// create an instance of tDSS by stored parameters from private key
-	std::stringstream dss_in;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl << dss_y << std::endl;
-	dss_in << dss_qual.size() << std::endl;
-	for (size_t i = 0; i < dss_qual.size(); i++)
-		dss_in << dss_qual[i] << std::endl;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl << dss_y << std::endl;
-	dss_in << dss_qual.size() << std::endl;
-	for (size_t i = 0; i < dss_qual.size(); i++)
-		dss_in << dss_qual[i] << std::endl;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl << dss_t << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl;
-	dss_in << "0" << std::endl << "0" << std::endl;
-	if (dss_x_rvss_qual.size())
-	{
-		// new private key format 107
-		dss_in << dss_x_rvss_qual.size() << std::endl;
-		for (size_t i = 0; i < dss_x_rvss_qual.size(); i++)
-			dss_in << dss_x_rvss_qual[i] << std::endl;
-	}
-	else
-	{
-		// old private key format 108
-		dss_in << dss_qual.size() << std::endl;
-		for (size_t i = 0; i < dss_qual.size(); i++)
-			dss_in << dss_qual[i] << std::endl;
-	}
-	assert((dss_c_ik.size() == dss_n));
-	for (size_t i = 0; i < dss_c_ik.size(); i++)
-	{
-		for (size_t j = 0; j < dss_c_ik.size(); j++)
-			dss_in << "0" << std::endl << "0" << std::endl;
-		assert((dss_c_ik[i].size() == (dss_t + 1)));
-		for (size_t k = 0; k < dss_c_ik[i].size(); k++)
-			dss_in << dss_c_ik[i][k] << std::endl;
-	}
-	if (opt_verbose)
-		std::cerr << "INFO: CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
-	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
-	if (!dss->CheckGroup())
-	{
-		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	
 	// reason for revocation
 	tmcg_openpgp_byte_t revcode = 0x00;
 	if (reason < 255)
 		revcode = reason;
 
 	// compute a hash of pub and sub, respectively
-	tmcg_openpgp_octets_t trailer_pub, pub_hashing, hash_pub, left_pub, trailer_sub, sub_hashing, hash_sub, left_sub;
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareRevocationSignature(0x20, hashalgo, csigtime, revcode, "", keyid, trailer_pub);
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareRevocationSignature(0x28, hashalgo, csigtime, revcode, "", keyid, trailer_sub);
-	for (size_t i = 6; i < pub.size(); i++)
-		pub_hashing.push_back(pub[i]);
-	for (size_t i = 6; i < sub.size(); i++)
-		sub_hashing.push_back(sub[i]);
+	tmcg_openpgp_octets_t trailer_pub, pub_hashing, hash_pub, left_pub,
+		trailer_sub, sub_hashing, hash_sub, left_sub;
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		PacketSigPrepareRevocationSignature(TMCG_OPENPGP_SIGNATURE_KEY_REVOCATION,
+			hashalgo, csigtime, revcode, "", prv->pub->id, trailer_pub);
+	if (sub != NULL)
+	{
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			PacketSigPrepareRevocationSignature(TMCG_OPENPGP_SIGNATURE_SUBKEY_REVOCATION,
+				hashalgo, csigtime, revcode, "", prv->pub->id, trailer_sub);
+	}
 	// RFC 4880 ERRATA:
-	// Primary key revocation signatures (type 0x20) hash only the key being revoked.
-	// Subkey revocation signature (type 0x28) hash first the primary key and then the
-	// subkey being revoked.
-	CallasDonnerhackeFinneyShawThayerRFC4880::KeyHash(pub_hashing, trailer_pub, hashalgo, hash_pub, left_pub);
-	CallasDonnerhackeFinneyShawThayerRFC4880::KeyHash(pub_hashing, sub_hashing, trailer_sub, hashalgo, hash_sub, left_sub);
+	// Primary key revocation signatures (type 0x20) hash only the key being
+	// revoked. Subkey revocation signature (type 0x28) hash first the primary
+	// key and then the subkey being revoked.
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		KeyHash(prv->pub->pub_hashing, trailer_pub, hashalgo, hash_pub,
+			left_pub);
+	if (sub != NULL)
+	{
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			KeyHash(prv->pub->pub_hashing, sub->pub->sub_hashing, trailer_sub,
+				hashalgo, hash_sub, left_sub);
+	}
 
 	// sign the hashes
 	tmcg_openpgp_octets_t revsig_pub, revsig_sub;
@@ -318,193 +360,315 @@ void run_instance
 	size_t buflen = 0;
 	gcry_error_t ret;
 	memset(buffer, 0, sizeof(buffer));
-	for (size_t i = 0; ((i < hash_pub.size()) && (i < sizeof(buffer))); i++, buflen++)
-		buffer[i] = hash_pub[i];
+	for (size_t i = 0; ((i < hash_pub.size()) && (i < sizeof(buffer)));
+		i++, buflen++)
+			buffer[i] = hash_pub[i];
 	r = gcry_mpi_new(2048);
 	s = gcry_mpi_new(2048);
 	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
 	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 	if (ret)
 	{
-		std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed" <<
+			" for h" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	if (!mpz_set_gcry_mpi(h, dsa_m))
+	if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
+			" failed for dsa_m" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	gcry_mpi_release(h);
 	std::stringstream err_log_sign;
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": dss.Sign() on pub" << std::endl;
-	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, idx2dkg, dkg2idx, aiou, rbc, err_log_sign))
+		std::cerr << "INFO: R_" << whoami << ": dss.Sign() on" <<
+			" primary key" << std::endl;
+	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, prv->tdss_idx2dkg,
+		prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on pub failed" << std::endl;
-		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
+		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on" <<
+			" primary key failed" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl <<
+			err_log_sign.str();
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-	if (!mpz_get_gcry_mpi(r, dsa_r))
+		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl <<
+			err_log_sign.str();
+	if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+			" failed for dsa_r" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(s, dsa_s))
+	if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
 	{
-		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
+		std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+			" failed for dsa_s" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_pub, left_pub, r, s, revsig_pub);
-	buflen = 0;
-	memset(buffer, 0, sizeof(buffer));
-	for (size_t i = 0; ((i < hash_sub.size()) && (i < sizeof(buffer))); i++, buflen++)
-		buffer[i] = hash_sub[i];
-	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
-	if (ret)
+	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_pub,
+		left_pub, r, s, revsig_pub);
+	if (sub != NULL)
 	{
-		std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
+		buflen = 0;
+		memset(buffer, 0, sizeof(buffer));
+		for (size_t i = 0; ((i < hash_sub.size()) && (i < sizeof(buffer)));
+			i++, buflen++)
+				buffer[i] = hash_sub[i];
+		ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
+		if (ret)
+		{
+			std::cerr << "ERROR: R_" << whoami << ": gcry_mpi_scan() failed" <<
+				" for h" << std::endl;
+			gcry_mpi_release(r), gcry_mpi_release(s);
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
+		if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
+		{
+			std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
+				" failed for dsa_m" << std::endl;
+			gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
+		gcry_mpi_release(h);
+		std::stringstream err_log_sign2;
+		if (opt_verbose)
+			std::cerr << "INFO: R_" << whoami << ": dss.Sign() on" <<
+				" first subkey" << std::endl;
+		if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s,
+			prv->tdss_idx2dkg, prv->tdss_dkg2idx, aiou, rbc, err_log_sign2))
+		{
+			std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on" <<
+				" first subkey failed" << std::endl;
+			std::cerr << "ERROR: R_" << whoami << ": log follows " <<
+				std::endl << err_log_sign2.str();
+			gcry_mpi_release(r), gcry_mpi_release(s);
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
+		if (opt_verbose > 1)
+			std::cerr << "INFO: R_" << whoami << ": log follows " <<
+				std::endl << err_log_sign2.str();
+		if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
+		{
+			std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+				" failed for dsa_r" << std::endl;
+			gcry_mpi_release(r), gcry_mpi_release(s);
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
+		if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
+		{
+			std::cerr << "ERROR: R_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+				" failed for dsa_s" << std::endl;
+			gcry_mpi_release(r), gcry_mpi_release(s);
+			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
+		CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_sub,
+			left_sub, r, s, revsig_sub);
 	}
-	if (!mpz_set_gcry_mpi(h, dsa_m))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	gcry_mpi_release(h);
-	std::stringstream err_log_sign2;
-	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": dss.Sign() on sub" << std::endl;
-	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, idx2dkg, dkg2idx, aiou, rbc, err_log_sign2))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": " << "tDSS Sign() on sub failed" << std::endl;
-		std::cerr << "ERROR: R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	if (opt_verbose > 1)
-		std::cerr << "INFO: R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
-	if (!mpz_get_gcry_mpi(r, dsa_r))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	if (!mpz_get_gcry_mpi(s, dsa_s))
-	{
-		std::cerr << "ERROR: R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
-		gcry_mpi_release(r), gcry_mpi_release(s);
-		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_sub, left_sub, r, s, revsig_sub);
 	gcry_mpi_release(r), gcry_mpi_release(s);
 	mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 
 	// at the end: deliver some more rounds for still waiting parties
-	time_t synctime = aiounicast::aio_timeout_very_long;
+	time_t synctime = aiounicast::aio_timeout_long;
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: R_" << whoami << ": waiting approximately " <<
+			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
+			std::endl;
 	rbc->Sync(synctime);
 
 	// export updated public key in OpenPGP armor format
 	std::stringstream pubfilename;
 	pubfilename << peers[whoami] << "_dkg-pub.asc";
 	std::string armor = "";
-	tmcg_openpgp_octets_t all, uid;
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketUidEncode(userid, uid);
-	all.insert(all.end(), pub.begin(), pub.end());
+	tmcg_openpgp_octets_t all;
+	all.insert(all.end(),
+		prv->pub->packet.begin(), prv->pub->packet.end());
 	all.insert(all.end(), revsig_pub.begin(), revsig_pub.end());
-	all.insert(all.end(), uid.begin(), uid.end());
-	all.insert(all.end(), uidsig.begin(), uidsig.end());
-	all.insert(all.end(), sub.begin(), sub.end());
-	all.insert(all.end(), revsig_sub.begin(), revsig_sub.end());
-	all.insert(all.end(), subsig.begin(), subsig.end());
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
+	for (size_t k = 0; k < prv->pub->selfsigs.size(); k++)
+		all.insert(all.end(),
+			(prv->pub->selfsigs[k]->packet).begin(),
+			(prv->pub->selfsigs[k]->packet).end());
+	for (size_t k = 0; k < prv->pub->keyrevsigs.size(); k++)
+		all.insert(all.end(),
+			(prv->pub->keyrevsigs[k]->packet).begin(),
+			(prv->pub->keyrevsigs[k]->packet).end());
+	for (size_t j = 0; j < prv->pub->userids.size(); j++)
+	{
+		TMCG_OpenPGP_UserID *uid = prv->pub->userids[j];
+		if (uid->valid)
+		{
+			all.insert(all.end(),
+				(uid->packet).begin(), (uid->packet).end());
+			for (size_t k = 0; k < uid->selfsigs.size(); k++)
+				all.insert(all.end(),
+					(uid->selfsigs[k]->packet).begin(),
+					(uid->selfsigs[k]->packet).end());
+			for (size_t k = 0; k < uid->revsigs.size(); k++)
+				all.insert(all.end(),
+					(uid->selfsigs[k]->packet).begin(),
+					(uid->selfsigs[k]->packet).end());
+		}
+	}
+	for (size_t j = 0; j < prv->pub->userattributes.size(); j++)
+	{
+		TMCG_OpenPGP_UserAttribute *uat = prv->pub->userattributes[j];
+		if (uat->valid)
+		{
+			all.insert(all.end(),
+				(uat->packet).begin(), (uat->packet).end());
+			for (size_t k = 0; k < uat->selfsigs.size(); k++)
+				all.insert(all.end(),
+					(uat->selfsigs[k]->packet).begin(),
+					(uat->selfsigs[k]->packet).end());
+			for (size_t k = 0; k < uat->revsigs.size(); k++)
+				all.insert(all.end(),
+					(uat->selfsigs[k]->packet).begin(),
+					(uat->selfsigs[k]->packet).end());
+		}
+	}
+	if (sub != NULL)
+	{
+		TMCG_OpenPGP_Subkey *psub = sub->pub;
+		all.insert(all.end(),
+			psub->packet.begin(), psub->packet.end());
+		all.insert(all.end(), revsig_sub.begin(), revsig_sub.end());
+		for (size_t k = 0; k < psub->selfsigs.size(); k++)
+			all.insert(all.end(),
+				(psub->selfsigs[k]->packet).begin(),
+				(psub->selfsigs[k]->packet).end());
+		for (size_t k = 0; k < psub->bindsigs.size(); k++)
+			all.insert(all.end(),
+				(psub->bindsigs[k]->packet).begin(),
+				(psub->bindsigs[k]->packet).end());
+		for (size_t k = 0; k < psub->pbindsigs.size(); k++)
+			all.insert(all.end(),
+				(psub->pbindsigs[k]->packet).begin(),
+				(psub->pbindsigs[k]->packet).end());
+		for (size_t k = 0; k < psub->keyrevsigs.size(); k++)
+			all.insert(all.end(),
+				(psub->keyrevsigs[k]->packet).begin(),
+				(psub->keyrevsigs[k]->packet).end());
+	}
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		ArmorEncode(TMCG_OPENPGP_ARMOR_PUBLIC_KEY_BLOCK, all, armor);
 	if (opt_verbose > 1)
 		std::cout << armor << std::endl;
 	std::ofstream pubofs((pubfilename.str()).c_str(), std::ofstream::out);
 	if (!pubofs.good())
 	{
-		std::cerr << "ERROR: R_" << whoami << ": opening public key file failed" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
+		std::cerr << "ERROR: R_" << whoami << ": opening public key file" <<
+			" failed" << std::endl;
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	pubofs << armor;
 	if (!pubofs.good())
 	{
-		std::cerr << "ERROR: R_" << whoami << ": writing public key file failed" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
+		std::cerr << "ERROR: R_" << whoami << ": writing public key file" <<
+			" failed" << std::endl;
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	pubofs.close();
 
-	// release tDSS
-	delete dss;
-
 	// release RBC
 	delete rbc;
 	
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": aiou.numRead = " << aiou->numRead <<
-			" aiou.numWrite = " << aiou->numWrite << std::endl;
+	{
+		std::cerr << "INFO: R_" << whoami << ": unicast channels";
+		aiou->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cerr << "INFO: R_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
-			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
+	{
+		std::cerr << "INFO: R_" << whoami << ": broadcast channel";
+		aiou2->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
 
 	// release asynchronous unicast and broadcast
 	delete aiou, delete aiou2;
 
 	// release
-	release_mpis();
+	delete dss;
+	delete ring;
+	delete prv;
 }
 
 #ifdef GNUNET
 char *gnunet_opt_hostname = NULL;
 char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
+char *gnunet_opt_k = NULL;
 unsigned int gnunet_opt_reason = 0;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
@@ -559,6 +723,12 @@ int main
 			"hostname (e.g. onion address) of this peer within PEERS",
 			&gnunet_opt_hostname
 		),
+		GNUNET_GETOPT_option_string('k',
+			"keyring",
+			"FILENAME",
+			"use keyring FILENAME containing external revocation keys",
+			&gnunet_opt_k
+		),
 		GNUNET_GETOPT_option_logfile(&logfile),
 		GNUNET_GETOPT_option_loglevel(&loglev),
 		GNUNET_GETOPT_option_string('p',
@@ -623,6 +793,8 @@ int main
 		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
+	if (gnunet_opt_k != NULL)
+		opt_k = gnunet_opt_k;
 	if (gnunet_opt_W != opt_W)
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 #endif
@@ -632,8 +804,12 @@ int main
 	{
 		std::string arg = argv[i+1];
 		// ignore options
-		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) || (arg.find("-r") == 0) || (arg.find("-w") == 0) || (arg.find("-W") == 0) || 
-			(arg.find("-L") == 0) || (arg.find("-l") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0))
+		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) ||
+			(arg.find("-r") == 0) || (arg.find("-w") == 0) ||
+			(arg.find("-W") == 0) || (arg.find("-L") == 0) ||
+			(arg.find("-l") == 0) || (arg.find("-x") == 0) ||
+			(arg.find("-P") == 0) || (arg.find("-H") == 0) ||
+			(arg.find("-k") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) && (opt_hostname == NULL))
@@ -641,6 +817,12 @@ int main
 				hostname = argv[i+1];
 				opt_hostname = (char*)hostname.c_str();
 			}
+			if ((arg.find("-k") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_k == NULL))
+			{
+				kfilename = argv[i+1];
+				opt_k = (char*)kfilename.c_str();
+			}
 			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) && (opt_passwords == NULL))
 			{
 				passwords = argv[i+1];
@@ -664,6 +846,8 @@ int main
 				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
 				std::cout << "  -H STRING      hostname (e.g. onion address) of this peer within PEERS" << std::endl;
+				std::cout << "  -k FILENAME    use keyring FILENAME" <<
+					" containing external revocation keys" << std::endl;
 				std::cout << "  -p INTEGER     start port for built-in TCP/IP message exchange service" << std::endl;
 				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
 				std::cout << "  -r INTEGER     reason for revocation (OpenPGP machine-readable code)" << std::endl;
@@ -705,7 +889,7 @@ int main
 	peers.push_back("Test3");
 	peers.push_back("Test4");
 	opt_r = 3;
-	opt_verbose = 1;
+	opt_verbose = 2;
 #endif
 
 	// check command line arguments
@@ -751,7 +935,6 @@ int main
 		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
 		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
 		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
-		libgcrypt_secmem = true;
 	}
 
 	// initialize LibTMCG
@@ -770,6 +953,11 @@ int main
 	{
 		if (port.length())
 			opt_p = strtoul(port.c_str(), NULL, 10); // get start port from options
+		if ((opt_p < 1) || (opt_p > 65535))
+		{
+			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
+			return -1;
+		}
 		tcpip_init(hostname);
 		tcpip_bindports((uint16_t)opt_p, false);
 		tcpip_bindports((uint16_t)opt_p, true);
@@ -794,6 +982,12 @@ int main
 			"hostname (e.g. onion address) of this peer within PEERS",
 			&gnunet_opt_hostname
 		),
+		GNUNET_GETOPT_option_string('k',
+			"keyring",
+			"FILENAME",
+			"use keyring FILENAME containing external revocation keys",
+			&gnunet_opt_k
+		),
 		GNUNET_GETOPT_option_string('p',
 			"port",
 			"STRING",
