@@ -48,78 +48,131 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
-#include "dkg-common.hh"
 #include "dkg-io.hh"
+#include "dkg-common.hh"
 
-int 					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-pid_t 					pid[DKGPG_MAX_N];
-std::vector<std::string>		peers;
-bool					instance_forked = false;
+int 						pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+int							broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+pid_t 						pid[DKGPG_MAX_N];
+std::vector<std::string>	peers;
+bool						instance_forked = false;
 
-std::string				passphrase, userid, ifilename, ofilename, passwords, hostname, port, URI;
-tmcg_openpgp_octets_t			keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
-std::map<size_t, size_t>		idx2dkg, dkg2idx;
-mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
-size_t					dss_n, dss_t, dss_i;
-std::vector<size_t>			dss_qual, dss_x_rvss_qual;
-std::vector< std::vector<mpz_ptr> >	dss_c_ik;
-mpz_t					dkg_p, dkg_q, dkg_g, dkg_h, dkg_x_i, dkg_xprime_i, dkg_y;
-size_t					dkg_n, dkg_t, dkg_i;
-std::vector<size_t>			dkg_qual;
-std::vector<mpz_ptr>			dkg_v_i;
-std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
-gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
-gcry_mpi_t				dsa_r, dsa_s, elg_r, elg_s, rsa_n, rsa_e, rsa_md;
-gcry_mpi_t 				gk, myk, sig_r, sig_s;
-gcry_mpi_t				revdsa_r, revdsa_s, revelg_r, revelg_s, revrsa_md;
+std::string					passphrase, ifilename, ofilename, kfilename;
+std::string					passwords, hostname, port, URI;
 
-int 					opt_verbose = 0;
-bool					libgcrypt_secmem = false;
-char					*opt_ifilename = NULL;
-char					*opt_ofilename = NULL;
-char					*opt_passwords = NULL;
-char					*opt_hostname = NULL;
-char					*opt_URI = NULL;
+int 						opt_verbose = 0;
+char						*opt_ifilename = NULL;
+char						*opt_ofilename = NULL;
+char						*opt_passwords = NULL;
+char						*opt_hostname = NULL;
+char						*opt_URI = NULL;
+char						*opt_k = NULL;
 unsigned long int			opt_e = 0, opt_p = 55000, opt_W = 5;
+bool						opt_t = false;
 
 void run_instance
-	(size_t whoami, const time_t sigtime, const time_t sigexptime, const size_t num_xtests)
+	(size_t whoami, const time_t sigtime, const time_t sigexptime,
+	 const size_t num_xtests)
 {
-	// read and parse the private key
+	// read the key file
 	std::string armored_seckey, thispeer = peers[whoami];
 	if (!check_strict_permissions(thispeer + "_dkg-sec.asc"))
 	{
-		std::cerr << "WARNING: weak permissions of private key file detected" << std::endl;
+		std::cerr << "WARNING: weak permissions of private key file" <<
+			" detected" << std::endl;
 		if (!set_strict_permissions(thispeer + "_dkg-sec.asc"))
 			exit(-1);
 	}
 	if (!read_key_file(thispeer + "_dkg-sec.asc", armored_seckey))
 		exit(-1);
-	init_mpis();
-	std::vector<std::string> CAPL;
-	time_t ckeytime = 0, ekeytime = 0;
-	if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
+
+	// read the keyring
+	std::string armored_pubring;
+	if (opt_k)
+	{
+		if (!read_key_file(kfilename, armored_pubring))
+			exit(-1);
+	}
+
+	// parse the keyring, the private key and corresponding signatures
+	TMCG_OpenPGP_Prvkey *prv = NULL;
+	TMCG_OpenPGP_Keyring *ring = NULL;
+	bool parse_ok;
+	if (opt_k)
+	{
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PublicKeyringParse(armored_pubring, opt_verbose, ring);
+		if (!parse_ok)
+		{
+			std::cerr << "WARNING: cannot use the given keyring" << std::endl;
+			ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
+		}
+	}
+	else
+		ring = new TMCG_OpenPGP_Keyring(); // create an empty keyring
+	parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+		PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
+	if (!parse_ok)
 	{
-		release_mpis();
-		keyid.clear(), subkeyid.clear(), pub.clear(), sub.clear(), uidsig.clear(), subsig.clear();
-		dss_qual.clear(), dss_x_rvss_qual.clear(), dss_c_ik.clear(), dkg_qual.clear(), dkg_v_i.clear(), dkg_c_ik.clear();
-		init_mpis();
-		// protected with password
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		if (!get_passphrase("Enter passphrase to unlock private key",
+			passphrase))
 		{
-			release_mpis();
+			std::cerr << "ERROR: cannot read passphrase" << std::endl;
+			delete ring;
 			exit(-1);
 		}
 #endif
-		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
-		{
-			std::cerr << "ERROR: S_" << whoami << ": wrong passphrase to unlock private key" << std::endl;
-			release_mpis();
-			exit(-1);
-		}
+		parse_ok = CallasDonnerhackeFinneyShawThayerRFC4880::
+			PrivateKeyBlockParse(armored_seckey, opt_verbose, passphrase, prv);
+	}
+	if (parse_ok)
+	{
+		prv->RelinkPublicSubkeys(); // relink the contained subkeys
+		prv->pub->CheckSelfSignatures(ring, opt_verbose);
+		prv->pub->CheckSubkeys(ring, opt_verbose);
+		prv->RelinkPrivateSubkeys(); // undo the relinking
+	}
+	else
+	{
+		std::cerr << "ERROR: cannot use the provided private key" << std::endl;
+		delete ring;
+		exit(-1);
+	}
+	if (!prv->pub->valid || prv->weak(opt_verbose))
+	{
+		std::cerr << "ERROR: primary key is invalid or weak" << std::endl;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	if (prv->pkalgo != TMCG_OPENPGP_PKALGO_EXPERIMENTAL7)
+	{
+		std::cerr << "ERROR: primary key is not a tDSS/DSA key" << std::endl;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+
+	// create an instance of tDSS by stored parameters from private key
+	CanettiGennaroJareckiKrawczykRabinDSS *dss = NULL;
+	if (!init_tDSS(prv, opt_verbose, dss))
+	{
+		delete dss;
+		delete ring;
+		delete prv;
+		exit(-1);
+	}
+	// create one-to-one mapping based on the stored canonicalized peer list
+	if (!prv->tDSS_CreateMapping(peers, opt_verbose))
+	{
+		std::cerr << "ERROR: creating 1-to-1 CAPL mapping failed" << std::endl;
+		delete dss;
+		delete ring;
+		delete prv;
+		exit(-1);
 	}
 
 	// create communication handles between all players
@@ -133,20 +186,31 @@ void run_instance
 			std::string pwd;
 			if (!TMCG_ParseHelper::gs(passwords, '/', pwd))
 			{
-				std::cerr << "ERROR: S_" << whoami << ": " << "cannot read password for protecting channel to S_" << i << std::endl;
-				release_mpis();
+				std::cerr << "ERROR: S_" << whoami << ": " << "cannot read" <<
+					" password for protecting channel to S_" << i << std::endl;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1);
 			}
 			key << pwd;
-			if (((i + 1) < peers.size()) && !TMCG_ParseHelper::nx(passwords, '/'))
+			if (((i + 1) < peers.size()) &&
+				!TMCG_ParseHelper::nx(passwords, '/'))
 			{
-				std::cerr << "ERROR: S_" << whoami << ": " << "cannot skip to next password for protecting channel to S_" << (i + 1) << std::endl;
-				release_mpis();
+				std::cerr << "ERROR: S_" << whoami << ": " << "cannot skip" <<
+					" to next password for protecting channel to S_" <<
+					(i + 1) << std::endl;
+				delete dss;
+				delete ring;
+				delete prv;
 				exit(-1);
 			}
 		}
 		else
-			key << "dkg-sign::S_" << (i + whoami); // use simple key -- we assume that GNUnet will provide secure channels
+		{
+			// use simple key -- we assume that GNUnet provides secure channels
+			key << "dkg-sign::S_" << (i + whoami);
+		}
 		uP_in.push_back(pipefd[i][whoami][0]);
 		uP_out.push_back(pipefd[whoami][i][1]);
 		uP_key.push_back(key.str());
@@ -156,17 +220,24 @@ void run_instance
 	}
 
 	// create asynchronous authenticated unicast channels
-	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami,
+		uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin,
+		(opt_W * 60));
 
 	// create asynchronous authenticated unicast channels for broadcast protocol
-	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami,
+		bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin,
+		(opt_W * 60));
 			
 	// create an instance of a reliable broadcast protocol (RBC)
 	std::string myID = "dkg-sign|";
 	for (size_t i = 0; i < peers.size(); i++)
 		myID += peers[i] + "|";
-	size_t T_RBC = (peers.size() - 1) / 3; // assume maximum asynchronous t-resilience for RBC
-	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
+	// assume maximum asynchronous t-resilience for RBC
+	size_t T_RBC = (peers.size() - 1) / 3;
+	CachinKursawePetzoldShoupRBC *rbc =
+		new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2,
+			aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
@@ -189,7 +260,8 @@ void run_instance
 
 	// participants must agree on a common signature creation time (OpenPGP)
 	if (opt_verbose)
-		std::cerr << "INFO: agree on a signature creation time for OpenPGP" << std::endl;
+		std::cerr << "INFO: agree on a signature creation time for OpenPGP" <<
+			std::endl;
 	time_t csigtime = 0;
 	std::vector<time_t> tvs;
 	mpz_t mtv;
@@ -208,7 +280,8 @@ void run_instance
 			}
 			else
 			{
-				std::cerr << "WARNING: S_" << whoami << ": no signature creation time stamp received from S_" << i << std::endl;
+				std::cerr << "WARNING: S_" << whoami << ": no signature" <<
+					" creation time stamp received from S_" << i << std::endl;
 			}
 		}
 	}
@@ -216,94 +289,81 @@ void run_instance
 	std::sort(tvs.begin(), tvs.end());
 	if (tvs.size() < (peers.size() - T_RBC))
 	{
-		std::cerr << "ERROR: S_" << whoami << ": not enough timestamps received" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": not enough timestamps" <<
+			" received" << std::endl;
 		delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	csigtime = tvs[tvs.size()/2]; // use a median value as some kind of gentle agreement
+	// use a median value as some kind of gentle agreement
+	csigtime = tvs[tvs.size()/2];
 	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": canonicalized signature creation time = " << csigtime << std::endl;
+		std::cerr << "INFO: S_" << whoami << ": canonicalized signature" <<
+			" creation time = " << csigtime << std::endl;
 
 	// select hash algorithm for OpenPGP based on |q| (size in bit)
 	tmcg_openpgp_hashalgo_t hashalgo = TMCG_OPENPGP_HASHALGO_UNKNOWN;
-	if (mpz_sizeinbase(dss_q, 2L) == 256)
+	if (mpz_sizeinbase(dss->q, 2L) == 256)
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA256; // SHA256 (alg 8)
-	else if (mpz_sizeinbase(dss_q, 2L) == 384)
+	else if (mpz_sizeinbase(dss->q, 2L) == 384)
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA384; // SHA384 (alg 9)
-	else if (mpz_sizeinbase(dss_q, 2L) == 512)
+	else if (mpz_sizeinbase(dss->q, 2L) == 512)
 		hashalgo = TMCG_OPENPGP_HASHALGO_SHA512; // SHA512 (alg 10)
 	else
 	{
-		std::cerr << "ERROR: S_" << whoami << ": selecting hash algorithm failed for |q| = " << mpz_sizeinbase(dss_q, 2L) << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": selecting hash algorithm" <<
+			" failed for |q| = " << mpz_sizeinbase(dss->q, 2L) << std::endl;
 		delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 
 	// compute the hash of the input file
 	if (opt_verbose)
-		std::cerr << "INFO: hashing the input file \"" << opt_ifilename << "\"" << std::endl;
+		std::cerr << "INFO: hashing the input file \"" << opt_ifilename <<
+			"\"" << std::endl;
 	tmcg_openpgp_octets_t trailer, hash, left;
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigPrepareDetachedSignature(0x00, hashalgo, csigtime, sigexptime, URI, keyid, trailer);
-	if (!CallasDonnerhackeFinneyShawThayerRFC4880::BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
+	if (opt_t)
 	{
-		std::cerr << "ERROR: S_" << whoami << ": BinaryDocumentHash() failed; cannot process input file \"" << opt_ifilename << "\"" << std::endl;
-		delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
-	}
-
-	// create an instance of tDSS by stored parameters from private key
-	std::stringstream dss_in;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl << dss_y << std::endl;
-	dss_in << dss_qual.size() << std::endl;
-	for (size_t i = 0; i < dss_qual.size(); i++)
-		dss_in << dss_qual[i] << std::endl;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl << dss_y << std::endl;
-	dss_in << dss_qual.size() << std::endl;
-	for (size_t i = 0; i < dss_qual.size(); i++)
-		dss_in << dss_qual[i] << std::endl;
-	dss_in << dss_p << std::endl << dss_q << std::endl << dss_g << std::endl << dss_h << std::endl;
-	dss_in << dss_n << std::endl << dss_t << std::endl << dss_i << std::endl << dss_t << std::endl;
-	dss_in << dss_x_i << std::endl << dss_xprime_i << std::endl;
-	dss_in << "0" << std::endl << "0" << std::endl;
-	if (dss_x_rvss_qual.size())
-	{
-		// new private key format 107
-		dss_in << dss_x_rvss_qual.size() << std::endl;
-		for (size_t i = 0; i < dss_x_rvss_qual.size(); i++)
-			dss_in << dss_x_rvss_qual[i] << std::endl;
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			PacketSigPrepareDetachedSignature(
+				TMCG_OPENPGP_SIGNATURE_CANONICAL_TEXT_DOCUMENT,
+				hashalgo, csigtime, sigexptime, URI, prv->pub->id, trailer);
+		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
+			TextDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
+		{
+			std::cerr << "ERROR: S_" << whoami << ": TextDocumentHash()" <<
+				" failed; cannot process input file \"" << opt_ifilename <<
+				"\"" << std::endl;
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
 	}
 	else
 	{
-		// old private key format 108
-		dss_in << dss_qual.size() << std::endl;
-		for (size_t i = 0; i < dss_qual.size(); i++)
-			dss_in << dss_qual[i] << std::endl;
-	}
-	assert((dss_c_ik.size() == dss_n));
-	for (size_t i = 0; i < dss_c_ik.size(); i++)
-	{
-		for (size_t j = 0; j < dss_c_ik.size(); j++)
-			dss_in << "0" << std::endl << "0" << std::endl;
-		assert((dss_c_ik[i].size() == (dss_t + 1)));
-		for (size_t k = 0; k < dss_c_ik[i].size(); k++)
-			dss_in << dss_c_ik[i][k] << std::endl;
-	}
-	if (opt_verbose)
-		std::cerr << "INFO: CanettiGennaroJareckiKrawczykRabinDSS(in, ...)" << std::endl;
-	CanettiGennaroJareckiKrawczykRabinDSS *dss = new CanettiGennaroJareckiKrawczykRabinDSS(dss_in);
-	if (!dss->CheckGroup())
-	{
-		std::cerr << "ERROR: S_" << whoami << ": " << "tDSS domain parameters are not correctly generated!" << std::endl;
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
-		exit(-1);
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			PacketSigPrepareDetachedSignature(
+				TMCG_OPENPGP_SIGNATURE_BINARY_DOCUMENT,
+				hashalgo, csigtime, sigexptime, URI, prv->pub->id, trailer);
+		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
+			BinaryDocumentHash(opt_ifilename, trailer, hashalgo, hash, left))
+		{
+			std::cerr << "ERROR: S_" << whoami << ": BinaryDocumentHash()" <<
+				" failed; cannot process input file \"" << opt_ifilename <<
+				"\"" << std::endl;
+			delete rbc, delete aiou, delete aiou2;
+			delete dss;
+			delete ring;
+			delete prv;
+			exit(-1);
+		}
 	}
 
 	// sign the hash
@@ -313,102 +373,159 @@ void run_instance
 	size_t buflen = 0;
 	gcry_error_t ret;
 	memset(buffer, 0, sizeof(buffer));
-	for (size_t i = 0; ((i < hash.size()) && (i < sizeof(buffer))); i++, buflen++)
-		buffer[i] = hash[i];
+	if (opt_verbose)
+		std::cerr << std::hex << "INFO: hash = ";
+	for (size_t i = 0; i < hash.size(); i++, buflen++)
+	{
+		if (i < sizeof(buffer))
+			buffer[i] = hash[i];
+		if (opt_verbose)
+			std::cerr << (int)hash[i] << " ";
+	}
+	if (opt_verbose)
+		std::cerr << std::dec << std::endl;
 	r = gcry_mpi_new(2048);
 	s = gcry_mpi_new(2048);
 	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
 	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 	if (ret)
 	{
-		std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": gcry_mpi_scan() failed" <<
+			" for h" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	if (!mpz_set_gcry_mpi(h, dsa_m))
+	if (opt_verbose > 1)
+		std::cerr << "INFO: S_" << whoami << ": h = " << h << std::endl;
+	if (!tmcg_mpz_set_gcry_mpi(h, dsa_m))
 	{
-		std::cerr << "ERROR: S_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_set_gcry_mpi()" <<
+			" failed for dsa_m" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	gcry_mpi_release(h);
 	std::stringstream err_log_sign;
 	if (opt_verbose)
 		std::cerr << "INFO: S_" << whoami << ": dss.Sign()" << std::endl;
-	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, idx2dkg, dkg2idx, aiou, rbc, err_log_sign))
+	if (!dss->Sign(peers.size(), whoami, dsa_m, dsa_r, dsa_s, prv->tdss_idx2dkg,
+		prv->tdss_dkg2idx, aiou, rbc, err_log_sign))
 	{
-		std::cerr << "ERROR: S_" << whoami << ": " << "tDSS Sign() failed" << std::endl;
-		std::cerr << "ERROR: S_" << whoami << ": log follows " << std::endl << err_log_sign.str();
+		std::cerr << "ERROR: S_" << whoami << ": " << "tDSS Sign() failed" <<
+			std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": log follows " << std::endl <<
+			err_log_sign.str();
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	if (opt_verbose > 1)
-		std::cerr << "INFO: S_" << whoami << ": log follows " << std::endl << err_log_sign.str(); 
-	if (!mpz_get_gcry_mpi(r, dsa_r))
+		std::cerr << "INFO: S_" << whoami << ": log follows " << std::endl <<
+			err_log_sign.str(); 
+	if (!tmcg_mpz_get_gcry_mpi(r, dsa_r))
 	{
-		std::cerr << "ERROR: S_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+			" failed for dsa_r" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(s, dsa_s))
+	if (!tmcg_mpz_get_gcry_mpi(s, dsa_s))
 	{
-		std::cerr << "ERROR: S_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
+		std::cerr << "ERROR: S_" << whoami << ": tmcg_mpz_get_gcry_mpi()" <<
+			" failed for dsa_s" << std::endl;
 		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dss, delete rbc, delete aiou, delete aiou2;
-		release_mpis();
+		delete rbc, delete aiou, delete aiou2;
+		delete dss;
+		delete ring;
+		delete prv;
 		exit(-1);
 	}
 	tmcg_openpgp_octets_t sig;
-	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer, left, r, s, sig);
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		PacketSigEncode(trailer, left, r, s, sig);
 	gcry_mpi_release(r), gcry_mpi_release(s);
 	mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 
 	// at the end: deliver some more rounds for still waiting parties
-	time_t synctime = aiounicast::aio_timeout_very_long;
+	time_t synctime = aiounicast::aio_timeout_long;
 	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": waiting approximately " << (synctime * (T_RBC + 1)) << " seconds for stalled parties" << std::endl;
+		std::cerr << "INFO: S_" << whoami << ": waiting approximately " <<
+			(synctime * (T_RBC + 1)) << " seconds for stalled parties" <<
+			std::endl;
 	rbc->Sync(synctime);
 
-	// release tDSS
-	delete dss;
-
 	// release RBC
 	delete rbc;
 	
 	// release handles (unicast channel)
 	uP_in.clear(), uP_out.clear(), uP_key.clear();
 	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": aiou.numRead = " << aiou->numRead <<
-			" aiou.numWrite = " << aiou->numWrite << std::endl;
+	{
+		std::cerr << "INFO: S_" << whoami << ": unicast channels";
+		aiou->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
 
 	// release handles (broadcast channel)
 	bP_in.clear(), bP_out.clear(), bP_key.clear();
 	if (opt_verbose)
-		std::cerr << "INFO: S_" << whoami << ": aiou2.numRead = " << aiou2->numRead <<
-			" aiou2.numWrite = " << aiou2->numWrite << std::endl;
+	{
+		std::cerr << "INFO: S_" << whoami << ": broadcast channel";
+		aiou2->PrintStatistics(std::cerr);
+		std::cerr << std::endl;
+	}
 
 	// release asynchronous unicast and broadcast
 	delete aiou, delete aiou2;
 
 	// release
-	release_mpis();
+	delete dss;
+	delete ring;
+	delete prv;
 
 	// output the result
 	std::string sigstr;
-	CallasDonnerhackeFinneyShawThayerRFC4880::ArmorEncode(TMCG_OPENPGP_ARMOR_SIGNATURE, sig, sigstr);
+	CallasDonnerhackeFinneyShawThayerRFC4880::
+		ArmorEncode(TMCG_OPENPGP_ARMOR_SIGNATURE, sig, sigstr);
+	if (opt_t)
+	{
+		std::string ct_head = "-----BEGIN PGP SIGNED MESSAGE-----\r\n";
+		std::string ct_hash; // construct corresponding Hash Armor Header
+		CallasDonnerhackeFinneyShawThayerRFC4880::
+			AlgorithmHashTextName(hashalgo, ct_hash);
+		// additional blank line is not included into message digest
+		ct_hash = "Hash: " + ct_hash + "\r\n\r\n";
+		std::string ct_body;
+		if (!CallasDonnerhackeFinneyShawThayerRFC4880::
+			DashEscapeFile(opt_ifilename, ct_body))
+		{
+			std::cerr << "ERROR: S_" << whoami << ": DashEscapeFile()" <<
+				" failed; cannot process input file \"" << opt_ifilename <<
+				"\"" << std::endl;
+			exit(-1);
+		}
+		sigstr = ct_head + ct_hash + ct_body + "\r\n" + sigstr;
+	}
 	if (opt_ofilename != NULL)
 	{
 		if (!write_message(opt_ofilename, sigstr))
@@ -425,11 +542,13 @@ char *gnunet_opt_ofilename = NULL;
 char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 char *gnunet_opt_URI = NULL;
+char *gnunet_opt_k = NULL;
 unsigned int gnunet_opt_sigexptime = 0;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
 unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_verbose = 0;
+int gnunet_opt_text = 0;
 #endif
 
 void fork_instance
@@ -444,7 +563,8 @@ void fork_instance
 			/* BEGIN child code: participant S_i */
 			time_t sigtime = time(NULL);
 #ifdef GNUNET
-			run_instance(whoami, sigtime, gnunet_opt_sigexptime, gnunet_opt_xtests);
+			run_instance(whoami, sigtime, gnunet_opt_sigexptime,
+				gnunet_opt_xtests);
 #else
 			run_instance(whoami, sigtime, opt_e, 0);
 #endif
@@ -488,15 +608,21 @@ int main
 		GNUNET_GETOPT_option_string('i',
 			"input",
 			"FILENAME",
-			"create detached signature from FILENAME",
+			"create signature from FILENAME",
 			&gnunet_opt_ifilename
 		),
+		GNUNET_GETOPT_option_string('k',
+			"keyring",
+			"FILENAME",
+			"use keyring FILENAME containing external revocation keys",
+			&gnunet_opt_k
+		),
 		GNUNET_GETOPT_option_logfile(&logfile),
 		GNUNET_GETOPT_option_loglevel(&loglev),
 		GNUNET_GETOPT_option_string('o',
 			"output",
 			"FILENAME",
-			"write detached signature to FILENAME",
+			"write generated signature to FILENAME",
 			&gnunet_opt_ofilename
 		),
 		GNUNET_GETOPT_option_string('p',
@@ -511,6 +637,11 @@ int main
 			"exchanged passwords to protect private and broadcast channels",
 			&gnunet_opt_passwords
 		),
+		GNUNET_GETOPT_option_flag('t',
+			"text",
+			"create cleartext signature",
+			&gnunet_opt_text
+		),
 		GNUNET_GETOPT_option_string('U',
 			"URI",
 			"STRING",
@@ -545,8 +676,9 @@ int main
 	};
 	if (GNUNET_STRINGS_get_utf8_args(argc, argv, &argc, &argv) != GNUNET_OK)
 	{
-		std::cerr << "ERROR: GNUNET_STRINGS_get_utf8_args() failed" << std::endl;
-    		return -1;
+		std::cerr << "ERROR: GNUNET_STRINGS_get_utf8_args() failed" <<
+			std::endl;
+   		return -1;
 	}
 	if (GNUNET_GETOPT_run(usage, options, argc, argv) == GNUNET_SYSERR)
 	{
@@ -567,6 +699,8 @@ int main
 		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
+	if (gnunet_opt_k != NULL)
+		opt_k = gnunet_opt_k;
 	if (gnunet_opt_W != opt_W)
 		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 	if (gnunet_opt_URI != NULL)
@@ -578,71 +712,114 @@ int main
 	{
 		std::string arg = argv[i+1];
 		// ignore options
-		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) || (arg.find("-w") == 0) || (arg.find("-W") == 0) || 
-		    (arg.find("-L") == 0) || (arg.find("-l") == 0) || (arg.find("-i") == 0) || (arg.find("-o") == 0) || 
-		    (arg.find("-e") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0) ||
-		    (arg.find("-U") == 0))
+		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) ||
+			(arg.find("-w") == 0) || (arg.find("-W") == 0) || 
+		    (arg.find("-L") == 0) || (arg.find("-l") == 0) ||
+			(arg.find("-i") == 0) || (arg.find("-o") == 0) || 
+		    (arg.find("-e") == 0) || (arg.find("-x") == 0) ||
+			(arg.find("-P") == 0) || (arg.find("-H") == 0) ||
+		    (arg.find("-U") == 0) || (arg.find("-k") == 0))
 		{
 			size_t idx = ++i;
-			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) && (opt_ifilename == NULL))
+			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_ifilename == NULL))
 			{
 				ifilename = argv[i+1];
 				opt_ifilename = (char*)ifilename.c_str();
 			}
-			if ((arg.find("-o") == 0) && (idx < (size_t)(argc - 1)) && (opt_ofilename == NULL))
+			if ((arg.find("-o") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_ofilename == NULL))
 			{
 				ofilename = argv[i+1];
 				opt_ofilename = (char*)ofilename.c_str();
 			}
-			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) && (opt_hostname == NULL))
+			if ((arg.find("-k") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_k == NULL))
+			{
+				kfilename = argv[i+1];
+				opt_k = (char*)kfilename.c_str();
+			}
+			if ((arg.find("-H") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_hostname == NULL))
 			{
 				hostname = argv[i+1];
 				opt_hostname = (char*)hostname.c_str();
 			}
-			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) && (opt_passwords == NULL))
+			if ((arg.find("-P") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_passwords == NULL))
 			{
 				passwords = argv[i+1];
 				opt_passwords = (char*)passwords.c_str();
 			}
-			if ((arg.find("-U") == 0) && (idx < (size_t)(argc - 1)) && (opt_URI == NULL))
+			if ((arg.find("-U") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_URI == NULL))
 			{
 				URI = argv[i+1];
 				opt_URI = (char*)URI.c_str();
 			}
-			if ((arg.find("-e") == 0) && (idx < (size_t)(argc - 1)) && (opt_e == 0))
+			if ((arg.find("-e") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_e == 0))
+			{
 				opt_e = strtoul(argv[i+1], NULL, 10);
-			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) && (port.length() == 0))
+			}
+			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) &&
+				(port.length() == 0))
+			{
 				port = argv[i+1];
-			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) && (opt_W == 5))
+			}
+			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) &&
+				(opt_W == 5))
+			{
 				opt_W = strtoul(argv[i+1], NULL, 10);
+			}
 			continue;
 		}
-		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-V") == 0))
+		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) ||
+			(arg.find("-h") == 0) || (arg.find("-V") == 0) ||
+			(arg.find("-t") == 0))
 		{
 			if ((arg.find("-h") == 0) || (arg.find("--help") == 0))
 			{
 #ifndef GNUNET
 				std::cout << usage << std::endl;
 				std::cout << about << std::endl;
-				std::cout << "Arguments mandatory for long options are also mandatory for short options." << std::endl;
+				std::cout << "Arguments mandatory for long options are also" <<
+					" mandatory for short options." << std::endl;
 				std::cout << "  -h, --help     print this help" << std::endl;
-				std::cout << "  -e TIME        expiration time of generated signature in seconds" << std::endl;
-				std::cout << "  -H STRING      hostname (e.g. onion address) of this peer within PEERS" << std::endl;
-				std::cout << "  -i FILENAME    create detached signature from FILENAME" << std::endl;
-				std::cout << "  -o FILENAME    write detached signature to FILENAME" << std::endl;
-				std::cout << "  -p INTEGER     start port for built-in TCP/IP message exchange service" << std::endl;
-				std::cout << "  -P STRING      exchanged passwords to protect private and broadcast channels" << std::endl;
-				std::cout << "  -U STRING      policy URI tied to generated signatures" << std::endl;
-				std::cout << "  -v, --version  print the version number" << std::endl;
-				std::cout << "  -V, --verbose  turn on verbose output" << std::endl;
-				std::cout << "  -W TIME        timeout for point-to-point messages in minutes" << std::endl;
+				std::cout << "  -e TIME        expiration time of generated" <<
+					" signature in seconds" << std::endl;
+				std::cout << "  -H STRING      hostname (e.g. onion address)" <<
+					" of this peer within PEERS" << std::endl;
+				std::cout << "  -i FILENAME    create signature from" <<
+					" FILENAME" << std::endl;
+				std::cout << "  -k FILENAME    use keyring FILENAME" <<
+					" containing external revocation keys" << std::endl;
+				std::cout << "  -o FILENAME    write generated signature to" <<
+					" FILENAME" << std::endl;
+				std::cout << "  -p INTEGER     start port for built-in" <<
+					" TCP/IP message exchange service" << std::endl;
+				std::cout << "  -P STRING      exchanged passwords to" <<
+					" protect private and broadcast channels" << std::endl;
+				std::cout << "  -t, --text     create cleartext signature" <<
+					std::endl;
+				std::cout << "  -U STRING      policy URI tied to generated" <<
+					" signatures" << std::endl;
+				std::cout << "  -v, --version  print the version number" <<
+					std::endl;
+				std::cout << "  -V, --verbose  turn on verbose output" <<
+					std::endl;
+				std::cout << "  -W TIME        timeout for point-to-point" <<
+					" messages in minutes" << std::endl;
 #endif
 				return 0; // not continue
 			}
+			if ((arg.find("-t") == 0) || (arg.find("--text") == 0))
+				opt_t = true;
 			if ((arg.find("-v") == 0) || (arg.find("--version") == 0))
 			{
 #ifndef GNUNET
-				std::cout << "dkg-sign v" << version << " without GNUNET support" << std::endl;
+				std::cout << "dkg-sign v" << version <<
+					" without GNUNET support" << std::endl;
 #endif
 				return 0; // not continue
 			}
@@ -662,7 +839,8 @@ int main
 		}
 		else
 		{
-			std::cerr << "ERROR: peer identity \"" << arg << "\" too long" << std::endl;
+			std::cerr << "ERROR: peer identity \"" << arg << "\" too long" <<
+				std::endl;
 			return -1;
 		}
 	}
@@ -675,29 +853,33 @@ int main
 	ofilename = "Test1_output.sig";
 	opt_ofilename = (char*)ofilename.c_str();
 	URI = "https://savannah.nongnu.org/projects/dkgpg/";
-	opt_verbose = 1;
+	opt_verbose = 2;
 #endif
 
 	// check command line arguments
 	if ((opt_hostname != NULL) && (opt_passwords == NULL))
 	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		std::cerr << "ERROR: option \"-P\" required due to insecure network" <<
+			std::endl;
 		return -1;
 	}
 	if (opt_ifilename == NULL)
 	{
-		std::cerr << "ERROR: option -i required to specify an input file" << std::endl;
+		std::cerr << "ERROR: option \"-i\" required to specify an input file" <<
+			std::endl;
 		return -1;
 	}
 	if (peers.size() < 1)
 	{
-		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
+		std::cerr << "ERROR: no peers given as argument; usage: " <<
+			usage << std::endl;
 		return -1;
 	}
 
 	// canonicalize peer list
 	std::sort(peers.begin(), peers.end());
-	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
+	std::vector<std::string>::iterator it =
+		std::unique(peers.begin(), peers.end());
 	peers.resize(std::distance(peers.begin(), it));
 	if ((peers.size() < 3)  || (peers.size() > DKGPG_MAX_N))
 	{
@@ -714,11 +896,13 @@ int main
 	// lock memory
 	if (!lock_memory())
 	{
-		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required for memory protection" << std::endl;
+		std::cerr << "WARNING: locking memory failed; CAP_IPC_LOCK required" <<
+			" for memory protection" << std::endl;
 		// at least try to use libgcrypt's secure memory
 		if (!gcry_check_version(TMCG_LIBGCRYPT_VERSION))
 		{
-			std::cerr << "ERROR: libgcrypt version >= " << TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
+			std::cerr << "ERROR: libgcrypt version >= " <<
+				TMCG_LIBGCRYPT_VERSION << " required" << std::endl;
 			return -1;
 		}
 		gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);
@@ -726,7 +910,6 @@ int main
 		gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0);
 		gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
 		gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);
-		libgcrypt_secmem = true;
 	}
 
 	// initialize LibTMCG
@@ -736,16 +919,21 @@ int main
 		return -1;
 	}
 	if (opt_verbose)
-		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
-	
+		std::cerr << "INFO: using LibTMCG version " << version_libTMCG() <<
+			std::endl;
 	
 	// initialize return code
 	int ret = 0;
-	// create underlying point-to-point channels, if built-in TCP/IP service requested
+	// create underlying point-to-point channels, if built-in TCP/IP requested
 	if (opt_hostname != NULL)
 	{
 		if (port.length())
-			opt_p = strtoul(port.c_str(), NULL, 10); // get start port from options
+			opt_p = strtoul(port.c_str(), NULL, 10); // start port from options
+		if ((opt_p < 1) || (opt_p > 65535))
+		{
+			std::cerr << "ERROR: no valid TCP start port given" << std::endl;
+			return -1;
+		}
 		tcpip_init(hostname);
 		tcpip_bindports((uint16_t)opt_p, false);
 		tcpip_bindports((uint16_t)opt_p, true);
@@ -779,13 +967,19 @@ int main
 		GNUNET_GETOPT_option_string('i',
 			"input",
 			"FILENAME",
-			"create detached signature from FILENAME",
+			"create signature from FILENAME",
 			&gnunet_opt_ifilename
 		),
+		GNUNET_GETOPT_option_string('k',
+			"keyring",
+			"FILENAME",
+			"use keyring FILENAME containing external revocation keys",
+			&gnunet_opt_k
+		),
 		GNUNET_GETOPT_option_string('o',
 			"output",
 			"FILENAME",
-			"write detached signature to FILENAME",
+			"write generated signature to FILENAME",
 			&gnunet_opt_ofilename
 		),
 		GNUNET_GETOPT_option_string('p',
@@ -806,6 +1000,11 @@ int main
 			"policy URI tied to signature",
 			&gnunet_opt_URI
 		),
+		GNUNET_GETOPT_option_flag('t',
+			"text",
+			"create cleartext signature",
+			&gnunet_opt_text
+		),
 		GNUNET_GETOPT_option_flag('V',
 			"verbose",
 			"turn on verbose output",
@@ -831,17 +1030,20 @@ int main
 		),
 		GNUNET_GETOPT_OPTION_END
 	};
-	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run, argv[0]);
+	ret = GNUNET_PROGRAM_run(argc, argv, usage, about, myoptions, &gnunet_run,
+		argv[0]);
 	GNUNET_free((void *) argv);
 	if (ret == GNUNET_OK)
 		return 0;
 	else
 		return -1;
 #else
-	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
+	std::cerr << "WARNING: GNUnet CADET is required for the message exchange" <<
+		" of this program" << std::endl;
 #endif
 
-	std::cerr << "INFO: running local test with " << peers.size() << " participants" << std::endl;
+	std::cerr << "INFO: running local test with " << peers.size() <<
+		" participants" << std::endl;
 	// open pipes
 	for (size_t i = 0; i < peers.size(); i++)
 	{
@@ -873,15 +1075,18 @@ int main
 		{
 			std::cerr << "ERROR: protocol instance ";
 			if (WIFSIGNALED(wstatus))
-				std::cerr << pid[i] << " terminated by signal " << WTERMSIG(wstatus) << std::endl;
+				std::cerr << pid[i] << " terminated by signal " <<
+					WTERMSIG(wstatus) << std::endl;
 			if (WCOREDUMP(wstatus))
 				std::cerr << pid[i] << " dumped core" << std::endl;
-			ret = -1;
+			ret = -1; // fatal error
 		}
 		else if (WIFEXITED(wstatus))
 		{
 			if (opt_verbose)
-				std::cerr << "INFO: protocol instance " << pid[i] << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+				std::cerr << "INFO: protocol instance " << pid[i] <<
+					" terminated with exit status " << WEXITSTATUS(wstatus) <<
+					std::endl;
 			if (WEXITSTATUS(wstatus))
 				ret = -2; // error
 		}
@@ -889,8 +1094,11 @@ int main
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
 				perror("ERROR: dkg-sign (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
+			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
+				(close(broadcast_pipefd[i][j][1]) < 0))
+			{
 				perror("ERROR: dkg-sign (close)");
+			}
 		}
 	}
 	
