@@ -1,7 +1,7 @@
 /*******************************************************************************
    This file is part of Distributed Privacy Guard (DKGPG).
 
- Copyright (C) 2017  Heiko Stamer <HeikoStamer@gmx.net>
+ Copyright (C) 2017, 2018  Heiko Stamer <HeikoStamer@gmx.net>
 
    DKGPG is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,34 +24,45 @@
 #endif
 #include "dkg-tcpip-common.hh"
 
-extern int				pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-extern pid_t				pid[DKGPG_MAX_N];
-extern std::vector<std::string>		peers;
-extern bool				instance_forked;
-extern int				opt_verbose;
-extern void				fork_instance(const size_t whoami);
+extern int						pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+extern int						broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+extern pid_t					pid[DKGPG_MAX_N];
+extern std::vector<std::string>	peers;
+extern bool						instance_forked;
+extern int						opt_verbose;
+extern void						fork_instance(const size_t whoami);
 
-static const size_t			tcpip_pipe_buffer_size = 4096;
+static const size_t				tcpip_pipe_buffer_size = 4096;
 
-std::string 				tcpip_thispeer;
-std::map<std::string, size_t> 		tcpip_peer2pipe;
-std::map<size_t, std::string> 		tcpip_pipe2peer;
-std::map<size_t, int>	 		tcpip_pipe2socket, tcpip_broadcast_pipe2socket;
-std::map<size_t, int>	 		tcpip_pipe2socket_out, tcpip_broadcast_pipe2socket_out;
-std::map<size_t, int>	 		tcpip_pipe2socket_in, tcpip_broadcast_pipe2socket_in;
+std::string 					tcpip_thispeer;
+std::map<std::string, size_t> 	tcpip_peer2pipe;
+std::map<size_t, std::string> 	tcpip_pipe2peer;
+std::map<size_t, int>	 		tcpip_pipe2socket;
+std::map<size_t, int>			tcpip_broadcast_pipe2socket;
+std::map<size_t, int>	 		tcpip_pipe2socket_out;
+std::map<size_t, int>			tcpip_broadcast_pipe2socket_out;
+std::map<size_t, int>	 		tcpip_pipe2socket_in;
+std::map<size_t, int>			tcpip_broadcast_pipe2socket_in;
 
-// This is the signal handler called when receiving SIGINT, SIGQUIT, and SIGTERM, respectively.
+typedef std::map<size_t, int>::const_iterator tcpip_mci_t;
+
+// This is the signal handler called when receiving SIGINT, SIGQUIT,
+// and SIGTERM, respectively.
 RETSIGTYPE tcpip_sig_handler_quit(int sig)
 {
-	if (instance_forked && (pid[tcpip_peer2pipe[tcpip_thispeer]] == 0)) // child process?
+	// child process?
+	if (instance_forked && (pid[tcpip_peer2pipe[tcpip_thispeer]] == 0))
 	{
 		if (opt_verbose)
-			std::cerr << "tcpip_sig_handler_quit(): child got signal " << sig << std::endl;
+			std::cerr << "tcpip_sig_handler_quit(): child got signal " <<
+				sig << std::endl;
 	}
 	else
 	{
 		if (opt_verbose)
-			std::cerr << "tcpip_sig_handler_quit(): parent got signal " << sig << std::endl;
+			std::cerr << "tcpip_sig_handler_quit(): parent got signal " <<
+				sig << std::endl;
+		// remove our own signal handler and quit
 		signal(SIGINT, SIG_DFL);
 		signal(SIGQUIT, SIG_DFL);
 		signal(SIGTERM, SIG_DFL);
@@ -68,10 +79,12 @@ void tcpip_init
 	tcpip_thispeer = hostname;
 	// initialize peer2pipe and pipe2peer mapping
 	if (opt_verbose)
-		std::cout << "INFO: using built-in TCP/IP service for message exchange instead of GNUnet CADET" << std::endl;
+		std::cerr << "INFO: using built-in TCP/IP service for message" <<
+			" exchange instead of GNUnet CADET" << std::endl;
 	if (std::find(peers.begin(), peers.end(), hostname) == peers.end())
 	{
-		std::cerr << "ERROR: cannot find hostname \"" << hostname << "\" of this peer within PEERS" << std::endl;
+		std::cerr << "ERROR: cannot find hostname \"" << hostname << "\" of" <<
+			" this peer within PEERS" << std::endl;
 		exit(-1);
 	}
 	for (size_t i = 0; i < peers.size(); i++)
@@ -86,12 +99,12 @@ void tcpip_init
 		{
 			if (pipe2(pipefd[i][j], O_NONBLOCK) < 0)
 			{
-				perror("dkg-tcpip-common (pipe)");
+				perror("ERROR: dkg-tcpip-common (pipe)");
 				exit(-1);
 			}
 			if (pipe2(broadcast_pipefd[i][j], O_NONBLOCK) < 0)
 			{
-				perror("dkg-tcpip-common (pipe)");
+				perror("ERROR: dkg-tcpip-common (pipe)");
 				exit(-1);
 			}
 		}
@@ -102,17 +115,17 @@ void tcpip_init
 	act.sa_handler = &tcpip_sig_handler_quit;
 	if (sigaction(SIGINT, &act, NULL) < 0)
 	{
-		perror("dkg-tcpip-common (sigaction)");
+		perror("ERROR: dkg-tcpip-common (sigaction)");
 		exit(-1);
 	}
 	if (sigaction(SIGQUIT, &act, NULL) < 0)
 	{
-		perror("dkg-tcpip-common (sigaction)");
+		perror("ERROR: dkg-tcpip-common (sigaction)");
 		exit(-1);
 	}
 	if (sigaction(SIGTERM, &act, NULL) < 0)
 	{
-		perror("dkg-tcpip-common (sigaction)");
+		perror("ERROR: dkg-tcpip-common (sigaction)");
 		exit(-1);
 	}
 }
@@ -120,11 +133,31 @@ void tcpip_init
 void tcpip_bindports
 	(const uint16_t start, const bool broadcast)
 {
-	uint16_t local_start = start + (tcpip_peer2pipe[tcpip_thispeer] * peers.size());
+	if (opt_verbose > 2)
+		std::cerr << "INFO: tcpip_bindports(" << start << ", " <<
+			(broadcast ? "true" : "false") << ") called" << std::endl;
+	uint16_t peers_size = 0;
+	if (peers.size() <= DKGPG_MAX_N)
+	{
+		peers_size = (uint16_t)peers.size();
+	}
+	else
+	{
+		std::cerr << "ERROR: too many peers defined" << std::endl;
+		tcpip_close();
+		tcpip_done();
+		exit(-1);
+	}
+	uint16_t peer_offset = tcpip_peer2pipe[tcpip_thispeer] * peers_size;
+	uint16_t local_start = start + peer_offset;
+	uint16_t local_end = local_start + peers_size;
 	size_t i = 0;
 	if (broadcast)
-		local_start += (peers.size() * peers.size());
-	for (uint16_t port = local_start; port < (local_start + (uint16_t)peers.size()); port++, i++)
+	{
+		local_start += peers_size * peers_size; // use different port range
+		local_end += peers_size * peers_size;
+	}
+	for (uint16_t port = local_start; port < local_end; port++, i++)
 	{
 		struct addrinfo hints = { 0 }, *res, *rp;
 		hints.ai_family = AF_INET;
@@ -137,7 +170,7 @@ void tcpip_bindports
 		{
 			std::cerr << "ERROR: resolving wildcard address failed: ";
 			if (ret == EAI_SYSTEM)
-				perror("dkg-tcpip-common (getaddrinfo)");
+				perror("ERROR: dkg-tcpip-common (getaddrinfo)");
 			else
 				std::cerr << gai_strerror(ret);
 			std::cerr << std::endl;
@@ -148,37 +181,40 @@ void tcpip_bindports
 		int sockfd = -1;
 		for (rp = res; rp != NULL; rp = rp->ai_next)
 		{
-			if ((sockfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol)) < 0)
+			if ((sockfd = socket(rp->ai_family, rp->ai_socktype,
+				rp->ai_protocol)) < 0)
 			{
-				perror("dkg-tcpip-common (socket)");
+				perror("WARNING: dkg-tcpip-common (socket)");
 				continue; // try next address
 			}
 			char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
 			memset(hbuf, 0, sizeof(hbuf));
 			memset(sbuf, 0, sizeof(sbuf));
-			if ((ret = getnameinfo(rp->ai_addr, rp->ai_addrlen, hbuf, sizeof(hbuf), sbuf, sizeof(sbuf), 
+			if ((ret = getnameinfo(rp->ai_addr, rp->ai_addrlen,
+				hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
 				NI_NUMERICHOST | NI_NUMERICSERV)) != 0)
 			{
 				std::cerr << "ERROR: resolving wildcard address failed: ";
 				if (ret == EAI_SYSTEM)
-					perror("dkg-tcpip-common (getnameinfo)");
+					perror("ERROR: dkg-tcpip-common (getnameinfo)");
 				else
 					std::cerr << gai_strerror(ret);
 				std::cerr << std::endl;
 				if (close(sockfd) < 0)
-					perror("dkg-tcpip-common (close)");
+					perror("ERROR: dkg-tcpip-common (close)");
 				freeaddrinfo(res);
 				tcpip_close();
 				tcpip_done();
 				exit(-1);
 			}
 			if (opt_verbose)
-				std::cout << "INFO: bind TCP/IP port " << port << " at address " << hbuf << std::endl;
+				std::cerr << "INFO: bind TCP/IP port " << port <<
+					" at address " << hbuf << std::endl;
 			if (bind(sockfd, rp->ai_addr, rp->ai_addrlen) < 0)
 			{
-				perror("dkg-tcpip-common (bind)");
+				perror("WARNING: dkg-tcpip-common (bind)");
 				if (close(sockfd) < 0)
-					perror("dkg-tcpip-common (close)");
+					perror("ERROR: dkg-tcpip-common (close)");
 				sockfd = -1;
 				continue; // try next address
 			}
@@ -187,16 +223,17 @@ void tcpip_bindports
 		freeaddrinfo(res);
 		if ((rp == NULL) || (sockfd < 0))
 		{
-			std::cerr << "ERROR: cannot bind TCP/IP port " << port << " for any valid IP address of this host" << std::endl;
+			std::cerr << "ERROR: cannot bind TCP/IP port " << port <<
+				" for any valid IP address of this host" << std::endl;
 			tcpip_close();
 			tcpip_done();
 			exit(-1);
 		}
 		else if (listen(sockfd, SOMAXCONN) < 0)
 		{
-			perror("dkg-tcpip-common (listen)");
+			perror("ERROR: dkg-tcpip-common (listen)");
 			if (close(sockfd) < 0)
-				perror("dkg-tcpip-common (close)");
+				perror("ERROR: dkg-tcpip-common (close)");
 			tcpip_close();
 			tcpip_done();
 			exit(-1);
@@ -211,25 +248,44 @@ void tcpip_bindports
 size_t tcpip_connect
 	(const uint16_t start, const bool broadcast)
 {
+	if (opt_verbose > 2)
+		std::cerr << "INFO: tcpip_connect(" << start << ", " <<
+			(broadcast ? "true" : "false") << ") called" << std::endl;
 	for (size_t i = 0; i < peers.size(); i++)
 	{
-		if ((broadcast && !tcpip_broadcast_pipe2socket_out.count(i)) || (!broadcast && !tcpip_pipe2socket_out.count(i)))
+		if ((broadcast && (tcpip_broadcast_pipe2socket_out.count(i) == 0)) ||
+			(!broadcast && (tcpip_pipe2socket_out.count(i) == 0)))
 		{
-			uint16_t port = start + (i * peers.size()) + (uint16_t)tcpip_peer2pipe[tcpip_thispeer];
+			uint16_t peers_size = 0;
+			if (peers.size() <= DKGPG_MAX_N)
+			{
+				peers_size = (uint16_t)peers.size();
+			}
+			else
+			{
+				std::cerr << "ERROR: too many peers defined" << std::endl;
+				tcpip_close();
+				tcpip_done();
+				exit(-1);
+			}
+			uint16_t peer_offset = (uint16_t)tcpip_peer2pipe[tcpip_thispeer];
+			uint16_t port = start + (i * peers_size) + peer_offset;
 			int ret;
 			struct addrinfo hints = { 0 }, *res, *rp;
 			hints.ai_family = AF_INET;
 			hints.ai_socktype = SOCK_STREAM;
 			hints.ai_flags = AI_NUMERICSERV | AI_ADDRCONFIG;
 			if (broadcast)
-				port += (peers.size() * peers.size());
+				port += peers_size * peers_size; // use different port range
 			std::stringstream ports;
 			ports << port;
-			if ((ret = getaddrinfo(peers[i].c_str(), (ports.str()).c_str(), &hints, &res)) != 0)
+			if ((ret = getaddrinfo(peers[i].c_str(), (ports.str()).c_str(),
+				&hints, &res)) != 0)
 			{
-				std::cerr << "ERROR: resolving hostname \"" << peers[i] << "\" failed: ";
+				std::cerr << "ERROR: resolving hostname \"" <<
+					peers[i] << "\" failed: ";
 				if (ret == EAI_SYSTEM)
-					perror("dkg-tcpip-common (getaddrinfo)");
+					perror("ERROR: dkg-tcpip-common (getaddrinfo)");
 				else
 					std::cerr << gai_strerror(ret);
 				std::cerr << std::endl;
@@ -240,17 +296,18 @@ size_t tcpip_connect
 			for (rp = res; rp != NULL; rp = rp->ai_next)
 			{
 				int sockfd;
-				if ((sockfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol)) < 0)
+				if ((sockfd = socket(rp->ai_family, rp->ai_socktype,
+					rp->ai_protocol)) < 0)
 				{
-					perror("dkg-tcpip-common (socket)");
+					perror("WARNING: dkg-tcpip-common (socket)");
 					continue; // try next address
 				}
 				if (connect(sockfd, rp->ai_addr, rp->ai_addrlen) < 0)
 				{
 					if (errno != ECONNREFUSED)
-						perror("dkg-tcpip-common (connect)");					
+						perror("WARNING: dkg-tcpip-common (connect)");					
 					if (close(sockfd) < 0)
-						perror("dkg-tcpip-common (close)");
+						perror("ERROR: dkg-tcpip-common (close)");
 					continue; // try next address
 				}
 				else
@@ -258,26 +315,30 @@ size_t tcpip_connect
 					char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
 					memset(hbuf, 0, sizeof(hbuf));
 					memset(sbuf, 0, sizeof(sbuf));
-					if ((ret = getnameinfo(rp->ai_addr, rp->ai_addrlen, hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
+					if ((ret = getnameinfo(rp->ai_addr, rp->ai_addrlen,
+						hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
 						NI_NUMERICHOST | NI_NUMERICSERV)) != 0)
 					{
-						std::cerr << "ERROR: resolving hostname \"" << peers[i] << "\" failed: ";
+						std::cerr << "ERROR: resolving hostname \"" <<
+							peers[i] << "\" failed: ";
 						if (ret == EAI_SYSTEM)
-							perror("dkg-tcpip-common (getnameinfo)");
+							perror("ERROR: dkg-tcpip-common (getnameinfo)");
 						else
 							std::cerr << gai_strerror(ret);
 						std::cerr << std::endl;
 						if (close(sockfd) < 0)
-							perror("dkg-tcpip-common (close)");
+							perror("ERROR: dkg-tcpip-common (close)");
 						freeaddrinfo(res);
 						tcpip_close();
 						tcpip_done();
 						exit(-1);
 					}
 					if (opt_verbose)
-						std::cout << "INFO: resolved hostname \"" << peers[i] << "\" to address " << hbuf << std::endl;
+						std::cerr << "INFO: resolved hostname \"" <<
+							peers[i] << "\" to address " << hbuf << std::endl;
 					if (opt_verbose)
-						std::cout << "INFO: connected to host \"" << peers[i] << "\" on port " << port << std::endl;
+						std::cerr << "INFO: connected to host \"" <<
+							peers[i] << "\" on port " << port << std::endl;
 					if (broadcast)
 						tcpip_broadcast_pipe2socket_out[i] = sockfd;
 					else
@@ -297,19 +358,24 @@ size_t tcpip_connect
 void tcpip_accept
 	()
 {
-	while ((tcpip_pipe2socket_in.size() < peers.size()) || (tcpip_broadcast_pipe2socket_in.size() < peers.size()))
+	if (opt_verbose > 2)
+		std::cerr << "INFO: tcpip_accept(...) called" << std::endl;
+	while ((tcpip_pipe2socket_in.size() < peers.size()) ||
+		(tcpip_broadcast_pipe2socket_in.size() < peers.size()))
 	{
 		fd_set rfds;
 		struct timeval tv;
 		int retval, maxfd = 0;
 		FD_ZERO(&rfds);
-		for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket.begin(); pi != tcpip_pipe2socket.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_pipe2socket.begin();
+			pi != tcpip_pipe2socket.end(); ++pi)
 		{
 			FD_SET(pi->second, &rfds);
 			if (pi->second > maxfd)
 				maxfd = pi->second;
 		}
-		for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket.begin(); pi != tcpip_broadcast_pipe2socket.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket.begin();
+			pi != tcpip_broadcast_pipe2socket.end(); ++pi)
 		{
 			FD_SET(pi->second, &rfds);
 			if (pi->second > maxfd)
@@ -323,12 +389,12 @@ void tcpip_accept
 			if ((errno == EAGAIN) || (errno == EINTR))
 			{
 				if (errno == EAGAIN)
-					perror("dkg-tcpip-common (select)");
+					perror("WARNING: dkg-tcpip-common (select)");
 				continue;
 			}
 			else
 			{
-				perror("dkg-tcpip-common (select)");
+				perror("ERROR: dkg-tcpip-common (select)");
 				tcpip_close();
 				tcpip_done();
 				exit(-1);
@@ -336,7 +402,8 @@ void tcpip_accept
 		}
 		if (retval == 0)
 			continue; // timeout
-		for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket.begin(); pi != tcpip_pipe2socket.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_pipe2socket.begin();
+			pi != tcpip_pipe2socket.end(); ++pi)
 		{
 			if (FD_ISSET(pi->second, &rfds))
 			{
@@ -345,7 +412,7 @@ void tcpip_accept
 				int connfd = accept(pi->second, (struct sockaddr*)&sin, &slen);
 				if (connfd < 0)
 				{
-					perror("dkg-tcpip-common (accept)");
+					perror("ERROR: dkg-tcpip-common (accept)");
 					tcpip_close();
 					tcpip_done();
 					exit(-1);
@@ -353,12 +420,12 @@ void tcpip_accept
 				tcpip_pipe2socket_in[pi->first] = connfd;
 				char ipaddr[INET6_ADDRSTRLEN];
 				int ret;
-				if ((ret = getnameinfo((struct sockaddr *)&sin, slen, ipaddr, sizeof(ipaddr), NULL, 0,
-					NI_NUMERICHOST)) != 0)
+				if ((ret = getnameinfo((struct sockaddr *)&sin, slen,
+					ipaddr, sizeof(ipaddr), NULL, 0, NI_NUMERICHOST)) != 0)
 				{
 					std::cerr << "ERROR: resolving incoming address failed: ";
 					if (ret == EAI_SYSTEM)
-						perror("dkg-tcpip-common (getnameinfo)");
+						perror("ERROR: dkg-tcpip-common (getnameinfo)");
 					else
 						std::cerr << gai_strerror(ret);
 					std::cerr << std::endl;
@@ -367,10 +434,12 @@ void tcpip_accept
 					exit(-1);
 				}
 				if (opt_verbose)
-					std::cout << "INFO: accept connection for P/D/R/S_" << pi->first << " from address " << ipaddr << std::endl;
+					std::cerr << "INFO: accept connection for P/D/R/S_" <<
+						pi->first << " from address " << ipaddr << std::endl;
 			}
 		}
-		for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket.begin(); pi != tcpip_broadcast_pipe2socket.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket.begin();
+			pi != tcpip_broadcast_pipe2socket.end(); ++pi)
 		{
 			if (FD_ISSET(pi->second, &rfds))
 			{
@@ -379,18 +448,18 @@ void tcpip_accept
 				int connfd = accept(pi->second, (struct sockaddr*)&sin, &slen);
 				if (connfd < 0)
 				{
-					perror("dkg-tcpip-common (accept)");
+					perror("ERROR: dkg-tcpip-common (accept)");
 					exit(-1);
 				}
 				tcpip_broadcast_pipe2socket_in[pi->first] = connfd;
 				char ipaddr[INET6_ADDRSTRLEN];
 				int ret;
-				if ((ret = getnameinfo((struct sockaddr *)&sin, slen, ipaddr, sizeof(ipaddr), NULL, 0,
-					NI_NUMERICHOST)) != 0)
+				if ((ret = getnameinfo((struct sockaddr *)&sin, slen,
+					ipaddr, sizeof(ipaddr), NULL, 0, NI_NUMERICHOST)) != 0)
 				{
 					std::cerr << "ERROR: resolving incoming address failed: ";
 					if (ret == EAI_SYSTEM)
-						perror("dkg-tcpip-common (getnameinfo)");
+						perror("ERROR: dkg-tcpip-common (getnameinfo)");
 					else
 						std::cerr << gai_strerror(ret);
 					std::cerr << std::endl;
@@ -399,7 +468,8 @@ void tcpip_accept
 					exit(-1);
 				}
 				if (opt_verbose)
-					std::cout << "INFO: accept broadcast connection for P/D/R/S_" << pi->first << " from address " << 
+					std::cerr << "INFO: accept broadcast connection for" <<
+						" P/D/R/S_" << pi->first << " from address " << 
 						ipaddr << std::endl;
 			}
 		}
@@ -410,11 +480,12 @@ void tcpip_accept
 void tcpip_fork
 	()
 {
-	if ((tcpip_pipe2socket_in.size() == peers.size()) && (tcpip_broadcast_pipe2socket_in.size() == peers.size()))
+	if ((tcpip_pipe2socket_in.size() == peers.size()) &&
+		(tcpip_broadcast_pipe2socket_in.size() == peers.size()))
 	{
 		// fork instance
 		if (opt_verbose)
-			std::cout << "INFO: forking the protocol instance ..." << std::endl;
+			std::cerr << "INFO: forking the protocol instance ..." << std::endl;
 		fork_instance(tcpip_peer2pipe[tcpip_thispeer]);
 	}
 	else
@@ -429,16 +500,17 @@ void tcpip_fork
 int tcpip_io
 	()
 {
+	size_t thisidx = tcpip_peer2pipe[tcpip_thispeer]; // index of this peer
 	while (1)
 	{
 		if (instance_forked)
 		{
 			// exit, if forked instance has terminated 
 			int wstatus = 0;
-			int thispid = pid[tcpip_peer2pipe[tcpip_thispeer]];
+			int thispid = pid[thisidx];
 			int ret = waitpid(thispid, &wstatus, WNOHANG);
 			if (ret < 0)
-				perror("dkg-tcpip-common (waitpid)");
+				perror("ERROR: dkg-tcpip-common (waitpid)");
 			else if (ret == thispid)
 			{
 				instance_forked = false;
@@ -446,7 +518,8 @@ int tcpip_io
 				{
 					std::cerr << "ERROR: protocol instance ";
 					if (WIFSIGNALED(wstatus))
-						std::cerr << thispid << " terminated by signal " << WTERMSIG(wstatus) << std::endl;
+						std::cerr << thispid << " terminated by signal " <<
+							WTERMSIG(wstatus) << std::endl;
 					if (WCOREDUMP(wstatus))
 						std::cerr << thispid << " dumped core" << std::endl;
 					return -1;
@@ -454,7 +527,9 @@ int tcpip_io
 				else if (WIFEXITED(wstatus))
 				{
 					if (opt_verbose)
-						std::cerr << "INFO: protocol instance " << thispid << " terminated with exit status " << WEXITSTATUS(wstatus) << std::endl;
+						std::cerr << "INFO: protocol instance " << thispid <<
+							" terminated with exit status " <<
+							WEXITSTATUS(wstatus) << std::endl;
 					return WEXITSTATUS(wstatus);
 				}
 				return 0;
@@ -464,18 +539,20 @@ int tcpip_io
 		struct timeval tv;
 		int retval, maxfd = 0;
 		FD_ZERO(&rfds);
-		for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket_in.begin(); pi != tcpip_pipe2socket_in.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_pipe2socket_in.begin();
+			pi != tcpip_pipe2socket_in.end(); ++pi)
 		{
-			if (pi->first != tcpip_peer2pipe[tcpip_thispeer])
+			if (pi->first != thisidx)
 			{
 				FD_SET(pi->second, &rfds);
 				if (pi->second > maxfd)
 					maxfd = pi->second;
 			}
 		}
-		for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket_in.begin(); pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
+			pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
 		{
-			if (pi->first != tcpip_peer2pipe[tcpip_thispeer])
+			if (pi->first != thisidx)
 			{
 				FD_SET(pi->second, &rfds);
 				if (pi->second > maxfd)
@@ -484,14 +561,14 @@ int tcpip_io
 		}
 		for (size_t i = 0; i < peers.size(); i++)
 		{
-			if (i != tcpip_peer2pipe[tcpip_thispeer])
+			if (i != thisidx)
 			{
-				FD_SET(pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0], &rfds);
-				if (pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0] > maxfd)
-					maxfd = pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0];
-				FD_SET(broadcast_pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0], &rfds);
-				if (broadcast_pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0] > maxfd)
-					maxfd = broadcast_pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0];
+				FD_SET(pipefd[thisidx][i][0], &rfds);
+				if (pipefd[thisidx][i][0] > maxfd)
+					maxfd = pipefd[thisidx][i][0];
+				FD_SET(broadcast_pipefd[thisidx][i][0], &rfds);
+				if (broadcast_pipefd[thisidx][i][0] > maxfd)
+					maxfd = broadcast_pipefd[thisidx][i][0];
 			}
 		}
 		tv.tv_sec = 1;
@@ -502,12 +579,12 @@ int tcpip_io
 			if ((errno == EAGAIN) || (errno == EINTR))
 			{
 				if (errno == EAGAIN)
-					perror("dkg-tcpip-common (select)");
+					perror("WARNING: dkg-tcpip-common (select)");
 				continue;
 			}
 			else
 			{
-				perror("dkg-tcpip-common (select)");
+				perror("ERROR: dkg-tcpip-common (select)");
 				tcpip_close();
 				tcpip_done();
 				exit(-1);
@@ -515,23 +592,27 @@ int tcpip_io
 		}
 		if (retval == 0)
 			continue; // timeout
-		for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket_in.begin(); pi != tcpip_pipe2socket_in.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_pipe2socket_in.begin();
+			pi != tcpip_pipe2socket_in.end(); ++pi)
 		{
-			if ((pi->first != tcpip_peer2pipe[tcpip_thispeer]) && FD_ISSET(pi->second, &rfds))
+			if ((pi->first != thisidx) && FD_ISSET(pi->second, &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
 				ssize_t len = read(pi->second, buf, sizeof(buf));
 				if (len < 0)
 				{
-					if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+					if ((errno == EWOULDBLOCK) || (errno == EINTR))
 					{
-						if (errno == EAGAIN)
-							perror("dkg-tcpip-common (read)");
+						continue;
+					}
+					else if (errno == EAGAIN)
+					{
+						perror("WARNING: dkg-tcpip-common (read)");
 						continue;
 					}
 					else
 					{
-						perror("dkg-tcpip-common (read)");
+						perror("ERROR: dkg-tcpip-common (read)");
 						tcpip_close();
 						tcpip_done();
 						exit(-1);
@@ -539,7 +620,8 @@ int tcpip_io
 				}
 				else if (len == 0)
 				{
-					std::cerr << "WARNING: connection collapsed for P/D/R/S_" << pi->first << std::endl;
+					std::cerr << "WARNING: connection collapsed for" <<
+						" P/D/R/S_" << pi->first << std::endl;
 					tcpip_pipe2socket_out.erase(pi->first);
 					tcpip_pipe2socket_in.erase(pi->first);
 					break;
@@ -547,27 +629,35 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
-						std::cout << "INFO: received " << len << " bytes on connection for P/D/R/S_" << 
-							pi->first << std::endl;
+						std::cerr << "INFO: received " << len << " bytes on " <<
+							"connection for P/D/R/S_" << pi->first << std::endl;
 					ssize_t wnum = 0;
 					do
 					{
-						ssize_t num = write(pipefd[pi->first][tcpip_peer2pipe[tcpip_thispeer]][1],
+						ssize_t num = write(pipefd[pi->first][thisidx][1],
 								buf + wnum, len - wnum);
 						if (num < 0)
 						{
-							if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
-								if (errno == EAGAIN)
-									perror("dkg-tcpip-common (write)");
 								if (opt_verbose)
-									std::cerr << "sleeping for write into pipe ..." << std::endl;
+									std::cerr << "INFO: sleeping for write" <<
+										" into pipe ..." << std::endl;
+								sleep(1);
+								continue;
+							}
+							else if (errno == EAGAIN)
+							{
+								perror("WARNING: dkg-tcpip-common (write)");
+								if (opt_verbose)
+									std::cerr << "INFO: sleeping for write" <<
+										" into pipe ..." << std::endl;
 								sleep(1);
 								continue;
 							}
 							else
 							{
-								perror("dkg-tcpip-common (write)");
+								perror("ERROR: dkg-tcpip-common (write)");
 								tcpip_close();
 								tcpip_done();
 								exit(-1);
@@ -580,23 +670,27 @@ int tcpip_io
 				}
 			}
 		}
-		for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket_in.begin(); pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
+		for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
+			pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
 		{
-			if ((pi->first != tcpip_peer2pipe[tcpip_thispeer]) && FD_ISSET(pi->second, &rfds))
+			if ((pi->first != thisidx) && FD_ISSET(pi->second, &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
 				ssize_t len = read(pi->second, buf, sizeof(buf));
 				if (len < 0)
 				{
-					if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+					if ((errno == EWOULDBLOCK) || (errno == EINTR))
 					{
-						if (errno == EAGAIN)
-							perror("dkg-tcpip-common (read)");
+						continue;
+					}
+					else if (errno == EAGAIN)
+					{
+						perror("WARNING: dkg-tcpip-common (read)");
 						continue;
 					}
 					else
 					{
-						perror("dkg-tcpip-common (read)");
+						perror("ERROR: dkg-tcpip-common (read)");
 						tcpip_close();
 						tcpip_done();
 						exit(-1);
@@ -604,7 +698,8 @@ int tcpip_io
 				}
 				else if (len == 0)
 				{
-					std::cerr << "WARNING: broadcast connection collapsed for P/D/R/S_" << pi->first << std::endl;
+					std::cerr << "WARNING: broadcast connection collapsed" <<
+						" for P/D/R/S_" << pi->first << std::endl;
 					tcpip_broadcast_pipe2socket_out.erase(pi->first);
 					tcpip_broadcast_pipe2socket_in.erase(pi->first);
 					break;
@@ -612,27 +707,37 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
-						std::cout << "INFO: received " << len << " bytes on broadcast connection for P/D/R/S_" << 
+						std::cerr << "INFO: received " << len << " bytes on" <<
+							" broadcast connection for P/D/R/S_" <<
 							pi->first << std::endl;
 					ssize_t wnum = 0;
 					do
 					{
-						ssize_t num = write(broadcast_pipefd[pi->first][tcpip_peer2pipe[tcpip_thispeer]][1],
-								buf + wnum, len - wnum);
+						ssize_t num = write(
+							broadcast_pipefd[pi->first][thisidx][1], buf + wnum,
+							len - wnum);
 						if (num < 0)
 						{
-							if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
-								if (errno == EAGAIN)
-									perror("dkg-tcpip-common (write)");
 								if (opt_verbose)
-									std::cerr << "sleeping for write into pipe ..." << std::endl;
+									std::cerr << "INFO: sleeping for write" <<
+										" into pipe ..." << std::endl;
+								sleep(1);
+								continue;
+							}
+							else if (errno == EAGAIN)
+							{
+								perror("WARNING: dkg-tcpip-common (write)");
+								if (opt_verbose)
+									std::cerr << "INFO: sleeping for write" <<
+										" into pipe ..." << std::endl;
 								sleep(1);
 								continue;
 							}
 							else
 							{
-								perror("dkg-tcpip-common (write)");
+								perror("ERROR: dkg-tcpip-common (write)");
 								tcpip_close();
 								tcpip_done();
 								exit(-1);
@@ -647,21 +752,24 @@ int tcpip_io
 		}
 		for (size_t i = 0; i < peers.size(); i++)
 		{
-			if ((i != tcpip_peer2pipe[tcpip_thispeer]) && FD_ISSET(pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0], &rfds))
+			if ((i != thisidx) && FD_ISSET(pipefd[thisidx][i][0], &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
-				ssize_t len = read(pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0], buf, sizeof(buf));
+				ssize_t len = read(pipefd[thisidx][i][0], buf, sizeof(buf));
 				if (len < 0)
 				{
-					if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+					if ((errno == EWOULDBLOCK) || (errno == EINTR))
 					{
-						if (errno == EAGAIN)
-							perror("dkg-tcpip-common (read)");
+						continue;
+					}
+					else if (errno == EAGAIN)
+					{
+						perror("WARNING: dkg-tcpip-common (read)");
 						continue;
 					}
 					else
 					{
-						perror("dkg-tcpip-common (read)");
+						perror("ERROR: dkg-tcpip-common (read)");
 						tcpip_close();
 						tcpip_done();
 						exit(-1);
@@ -675,32 +783,43 @@ int tcpip_io
 				else if (tcpip_pipe2socket_out.count(i))
 				{
 					if (opt_verbose > 1)
-						std::cout << "INFO: sending " << len << " bytes on connection to P/D/R/S_" << i << std::endl;
+						std::cerr << "INFO: sending " << len << " bytes on" <<
+							" connection to P/D/R/S_" << i << std::endl;
 					ssize_t wnum = 0;
 					do
 					{
-						ssize_t num = write(tcpip_pipe2socket_out[i], buf + wnum, len - wnum);
+						ssize_t num = write(tcpip_pipe2socket_out[i],
+							buf + wnum, len - wnum);
 						if (num < 0)
 						{
-							if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
-								if (errno == EAGAIN)
-									perror("dkg-tcpip-common (write)");
 								if (opt_verbose)
-									std::cerr << "sleeping for write into socket ..." << std::endl;
+									std::cerr << "INFO: sleeping for write" <<
+										" into socket ..." << std::endl;
+								sleep(1);
+								continue;
+							}
+							else if (errno == EAGAIN)
+							{
+								perror("WARNING: dkg-tcpip-common (write)");
+								if (opt_verbose)
+									std::cerr << "INFO: sleeping for write" <<
+										" into socket ..." << std::endl;
 								sleep(1);
 								continue;
 							}
 							else if (errno == ECONNRESET)
 							{
-								std::cerr << "WARNING: connection collapsed for P/D/R/S_" << i << std::endl;
+								std::cerr << "WARNING: connection collapsed" <<
+									" for P/D/R/S_" << i << std::endl;
 								tcpip_broadcast_pipe2socket_out.erase(i);
 								tcpip_broadcast_pipe2socket_in.erase(i);
 								break;
 							}
 							else
 							{
-								perror("dkg-tcpip-common (write)");
+								perror("ERROR: dkg-tcpip-common (write)");
 								tcpip_close();
 								tcpip_done();
 								exit(-1);
@@ -714,24 +833,30 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
-						std::cout << "INFO: discarding " << len << " bytes for P/D/R/S_" << i << std::endl;
+						std::cerr << "INFO: discarding " << len << " bytes" <<
+							" for P/D/R/S_" << i << std::endl;
 				}
 			}
-			if ((i != tcpip_peer2pipe[tcpip_thispeer]) && FD_ISSET(broadcast_pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0], &rfds))
+			if ((i != thisidx) &&
+				FD_ISSET(broadcast_pipefd[thisidx][i][0], &rfds))
 			{
 				char buf[tcpip_pipe_buffer_size];
-				ssize_t len = read(broadcast_pipefd[tcpip_peer2pipe[tcpip_thispeer]][i][0], buf, sizeof(buf));
+				ssize_t len = read(broadcast_pipefd[thisidx][i][0], buf,
+					sizeof(buf));
 				if (len < 0)
 				{
-					if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+					if ((errno == EWOULDBLOCK) || (errno == EINTR))
 					{
-						if (errno == EAGAIN)
-							perror("dkg-tcpip-common (read)");
+						continue;
+					}
+					else if (errno == EAGAIN)
+					{
+						perror("WARNING: dkg-tcpip-common (read)");
 						continue;
 					}
 					else
 					{
-						perror("dkg-tcpip-common (read)");
+						perror("ERROR: dkg-tcpip-common (read)");
 						tcpip_close();
 						tcpip_done();
 						exit(-1);
@@ -744,32 +869,44 @@ int tcpip_io
 				else if (tcpip_broadcast_pipe2socket_out.count(i))
 				{
 					if (opt_verbose > 1)
-						std::cout << "INFO: sending " << len << " bytes on broadcast connection to P/D/R/S_" << i << std::endl;
+						std::cerr << "INFO: sending " << len << " bytes on" <<
+							" broadcast connection to P/D/R/S_" << i <<
+							std::endl;
 					ssize_t wnum = 0;
 					do
 					{
-						ssize_t num = write(tcpip_broadcast_pipe2socket_out[i], buf + wnum, len - wnum);
+						ssize_t num = write(tcpip_broadcast_pipe2socket_out[i],
+							buf + wnum, len - wnum);
 						if (num < 0)
 						{
-							if ((errno == EAGAIN) || (errno == EWOULDBLOCK) || (errno == EINTR))
+							if ((errno == EWOULDBLOCK) || (errno == EINTR))
 							{
-								if (errno == EAGAIN)
-									perror("dkg-tcpip-common (write)");
 								if (opt_verbose)
-									std::cerr << "sleeping for write into socket ..." << std::endl;
+									std::cerr << "INFO: sleeping for write" <<
+										" into socket ..." << std::endl;
+								sleep(1);
+								continue;
+							}
+							else if (errno == EAGAIN)
+							{
+								perror("WARNING: dkg-tcpip-common (write)");
+								if (opt_verbose)
+									std::cerr << "INFO: sleeping for write" <<
+										" into socket ..." << std::endl;
 								sleep(1);
 								continue;
 							}
 							else if (errno == ECONNRESET)
 							{
-								std::cerr << "WARNING: broadcast connection collapsed for P/D/R/S_" << i << std::endl;
+								std::cerr << "WARNING: broadcast connection" <<
+									" collapsed for P/D/R/S_" << i << std::endl;
 								tcpip_broadcast_pipe2socket_out.erase(i);
 								tcpip_broadcast_pipe2socket_in.erase(i);
 								break;
 							}
 							else
 							{
-								perror("dkg-tcpip-common (write)");
+								perror("ERROR: dkg-tcpip-common (write)");
 								tcpip_close();
 								tcpip_done();
 								exit(-1);
@@ -783,7 +920,8 @@ int tcpip_io
 				else
 				{
 					if (opt_verbose > 1)
-						std::cout << "INFO: discarding " << len << " bytes for P/D/R/S_" << i << std::endl;
+						std::cerr << "INFO: discarding " << len << " bytes" <<
+							" for P/D/R/S_" << i << std::endl;
 				}
 			}
 		}
@@ -793,35 +931,41 @@ int tcpip_io
 void tcpip_close
 	()
 {
-	for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket_in.begin(); pi != tcpip_pipe2socket_in.end(); ++pi)
+	for (tcpip_mci_t pi = tcpip_pipe2socket_in.begin();
+		pi != tcpip_pipe2socket_in.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("dkg-tcpip-common (close)");
+			perror("ERROR: dkg-tcpip-common (close)");
 	}
-	for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket_out.begin(); pi != tcpip_pipe2socket_out.end(); ++pi)
+	for (tcpip_mci_t pi = tcpip_pipe2socket_out.begin();
+		pi != tcpip_pipe2socket_out.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("dkg-tcpip-common (close)");
+			perror("ERROR: dkg-tcpip-common (close)");
 	}
-	for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket_in.begin(); pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
+	for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_in.begin();
+		pi != tcpip_broadcast_pipe2socket_in.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("dkg-tcpip-common (close)");
+			perror("ERROR: dkg-tcpip-common (close)");
 	}
-	for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket_out.begin(); pi != tcpip_broadcast_pipe2socket_out.end(); ++pi)
+	for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket_out.begin();
+		pi != tcpip_broadcast_pipe2socket_out.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("dkg-tcpip-common (close)");
+			perror("ERROR: dkg-tcpip-common (close)");
 	}
-	for (std::map<size_t, int>::const_iterator pi = tcpip_pipe2socket.begin(); pi != tcpip_pipe2socket.end(); ++pi)
+	for (tcpip_mci_t pi = tcpip_pipe2socket.begin();
+		pi != tcpip_pipe2socket.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("dkg-tcpip-common (close)");
+			perror("ERROR: dkg-tcpip-common (close)");
 	}
-	for (std::map<size_t, int>::const_iterator pi = tcpip_broadcast_pipe2socket.begin(); pi != tcpip_broadcast_pipe2socket.end(); ++pi)
+	for (tcpip_mci_t pi = tcpip_broadcast_pipe2socket.begin();
+		pi != tcpip_broadcast_pipe2socket.end(); ++pi)
 	{
 		if (close(pi->second) < 0)
-			perror("dkg-tcpip-common (close)");
+			perror("ERROR: dkg-tcpip-common (close)");
 	}
 }
 
@@ -832,22 +976,24 @@ void tcpip_done
 	{
 		int thispid = pid[tcpip_peer2pipe[tcpip_thispeer]];
 		if (opt_verbose)
-			std::cout << "kill(" << thispid << ", SIGTERM)" << std::endl;
+			std::cerr << "INFO: kill(" << thispid << ", SIGTERM)" << std::endl;
 		if(kill(thispid, SIGTERM))
-			perror("dkg-tcpip-common (kill)");
+			perror("ERROR: dkg-tcpip-common (kill)");
 		if (opt_verbose)
-			std::cout << "waitpid(" << thispid << ", NULL, 0)" << std::endl;
+			std::cerr << "INFO: waitpid(" << thispid << ", NULL, 0)" <<
+				std::endl;
 		if (waitpid(thispid, NULL, 0) != thispid)
-			perror("dkg-tcpip-common (waitpid)");
+			perror("ERROR: dkg-tcpip-common (waitpid)");
 	}
 	for (size_t i = 0; i < peers.size(); i++)
 	{
 		for (size_t j = 0; j < peers.size(); j++)
 		{
 			if ((close(pipefd[i][j][0]) < 0) || (close(pipefd[i][j][1]) < 0))
-				perror("dkg-tcpip-common (close)");
-			if ((close(broadcast_pipefd[i][j][0]) < 0) || (close(broadcast_pipefd[i][j][1]) < 0))
-				perror("dkg-tcpip-common (close)");
+				perror("ERROR: dkg-tcpip-common (close)");
+			if ((close(broadcast_pipefd[i][j][0]) < 0) ||
+				(close(broadcast_pipefd[i][j][1]) < 0))
+				perror("ERROR: dkg-tcpip-common (close)");
 		}
 	}
 }
